from numpy import trapezoid,array,interp,log,log10,pi,logspace,argmax
from scipy.optimize import brentq,fminbound
from scipy.integrate import fixed_quad
from OutLines.__funcs__ import *
# hard-wired constants
global c,mH,mu,Msun,Lsun,yr,pi4,kpc
c    = 2.99792458e5 # km/s
mH   = 1.6733e-24 # g
mu   = 0.6 # mean molecular weight
Msun = 1.99e33 # g
Lsun = 3.9e33 # erg/s
yr   = 365.25*86400 # s/yr
kpc  = 3.086e21 # cm
pi4  = 4*pi

class Properties(object):
    '''
    Name:
        Properties

    Purpose:
        Inherit an OutLines Profile object and calculate outflow properties
        based on the characteristic parameters of the line profile. Results are
        stored as attributes and can be viewed using the \'print_props\' method.

    Arguments:
        :model (*OutLine object*): `OutLines` model profile for emission or
                        absorption

    Attributes:
        :props (*dict*): dictionary of the outflow properties
        :units (*dict*): dictionary of the outflow properties' units

    Methods:
        :update_params: updates parameters stored in the Properties object
        :calc_props: calculates outflow properties from parameters; called by
                        __init__ and update_params
        :print_props: prints outflow properties to the command line
    '''
    def __init__(self,Profile):
        # inherit the wind
        vars(self).update(vars(Profile))
        self.set_pars()
        self.calc_props()
    # print documentation
    def docs(self):
        print(self.__doc__)
    # sets parameters from the provided Profile model
    def set_pars(self):
        self.vinf = self.params['TerminalVelocity']*c
        self.beta = self.params['VelocityIndex']
        if self.settings['Geometry'] == 'Spherical':
            self.t0 = pi
            self.t1 = 0
        elif self.settings['Geometry'] == 'FilledCones':
            self.inc = self.params['Inclination']
            self.t0  = self.params['OpeningAngle']
            self.t1  = 0.
        elif self.settings['Geometry'] == 'HollowConesFixedCavity':
            self.inc = self.params['Inclination']
            self.t0  = self.params['OpeningAngle']
            self.t1  = self.params['OpeningAngle']-0.174533
        else:
            self.inc = self.params['Inclination']
            self.t0 = self.params['OpeningAngle']
            self.t1 = self.params['CavityAngle']
        if self.settings['FromRest']:
            self.vini = 0
        else:
            self.vini = self.params['LaunchVelocity']
        self.denpars = array([self.params[param] for param in self.npname])
    # options for updating the parameters
    def update_params(self,par_name,par_val):
        if hasattr(par_val,'__len__'):
            for name,val in zip(par_name,par_val):
                self.update_params(name,val)
        else:
            # if velocity > 1, convert to c units
            if 'Terminal' in par_name or 'Doppler' in par_name:
                if par_val > 1 :
                    par_val = par_val/2.99792458e5
            if 'Inclination' in par_name or 'Angle' in par_name:
                if par_val > 2*pi :
                    par_val = par_val*pi/180
            self.params[par_name] = par_val
            if self.settings['StaticComponent'] :
                if 'Static' in par_name or 'Doppler' in par_name:
                    self.statps[par_name] = par_val
                else:
                    self.outfps[par_name] = par_val
        self.set_pars()
        self.calc_props()
    # density profile
    def den(self,w1):
        return n[self.settings['DensityProfile']](w1,self.beta,self.vini,\
                        self.settings['VelocityField'],*self.denpars)
    # characteristic outflow velocity via Flury+ 2023
    def vel(self,x1):
        return v[self.settings['VelocityField']](x1,self.beta,self.vini)
    # outflow momentum density via Flury+ 2023
    def mom(self,x1):
        return self.den(self.vel(x1))*self.vel(x1)
    # calculate properties of the outflow
    def calc_props(self):
        # characteristic outflow radius
        xarr = logspace(0,2,2001)
        xout_guess = xarr[argmax(self.mom(xarr))]
        xout = fminbound(lambda x1: -self.mom(x1),1,2*xout_guess)
        # characteristic outflow velocity
        vout = self.vinf*self.vel(xout)
        # integrated density profile
        self.Rcal = lambda x1: self.den(self.vel(x1))
        Rcal = fixed_quad(self.Rcal,1,xout)[0]
        # mass outflow rate in km*Msun/yr (need to multiply by R0^2 n0)
        Mdot = pi4*(cos(self.t1)-cos(self.t0))*mu*mH*vout*Rcal*yr/Msun*kpc**2*1e5
        # relative momentum injection rate in dyne
        pdot = Mdot*vout*10**(30.7997-34)
        pdot_rel = ((cos(self.t1)-cos(self.t0))*Rcal*(vout/100)**2)
        #log10(pi4*(cos(self.t1)-cos(self.t0))*mu*mH*vout**2*Rcal * dyne)
        # relative energy injection rate in erg/s
        Edot =  (0.5*Mdot*vout**2) * 10**(35.7997-42)
        Edot_rel = ((cos(self.t1)-cos(self.t0))*Rcal*(vout/100)**3)
        self.props = {'x.out':xout,'v.out':vout,'Mdot':Mdot,'pdot':pdot,\
                      'Edot':Edot,'v.esc':vout/100,'pdot.esc':pdot_rel,\
                      'Edot.esc':Edot_rel}
        self.units = {'x.out':'','v.out':'km s^-1','Mdot':'Msun yr^-1 ',\
                    'pdot':'10^34 dyne','Edot':'10^42 erg s^-1',
                    'v.esc':'','pdot.esc':'','Edot.esc':''}
    # print outflow properties
    def print_props(self):
        print(3*' '+43*'-'+'\n  |'+13*' '+'MODEL PROPERTIES'+14*' '+'|\n'+3*' '+43*'-')
        for param,value in iter(self.props.items()):
            print(f'  | {param: >12s} : {value: >8.3f}  {self.units[param]: <16s} |')
        print(3*' '+43*'-')
        print('  | Mdot, pdot, Edot / R0^2 n0 [kpc^2 cm^-3]  |')
        print('  |  pdot.est, Edot.esc / v0 [100 km s^-1]    |\n'+3*' '+43*'-')



from OutLines.__funcs__ import *
# velocity integral limits
def calc_limits(ww0,vinf,vini,vapr,beta,VF,Inflow=False):
    # error check for length of w
    if not hasattr(ww0,'__len__'):
        w = array([ww0])
    # observed velocity from wavelengths, in vinf/c units
    y = absolute(((ww0)**2-1)/((ww0)**2+1)) / vinf
    ind = where(y<1)[0]
    # u = v/vinf (not y since y here is used for vobs/v)
    # with no occultation, the minimum velocity is just the observed velocity
    umin = y[ind]
    # primary cone type -- start with all as anterior
    cone = array(len(umin)*['ante']).astype(str)
    # when occultation by source occurs, need to solve
    # for minimum contributing velocity to observed y
    if Inflow :
        umin[ww0[ind]<1] = array(list(map(partial(solve_u0,beta,vini,VF),y[ind][ww0[ind]<1])))
        cone[ww0[ind]<1] = 'post'
    else:
        umin[ww0[ind]>1] = array(list(map(partial(solve_u0,beta,vini,VF),y[ind][ww0[ind]>1])))
        cone[ww0[ind]>1] = 'post'
    # by definition, velocity cannot exceed terminal velocity
    # since u = v/vinf, max value is simply 1 unless aperture effects occur
    umax = ones(len(ind))
    if vapr < vinf :
        wapr = array(list(map(partial(solve_u1,vapr/vinf,beta,vini,VF),y[ind])))
        umax = min([umax,wapr],axis=0)
    if vini > 0 :
        umin[umin<vini/vinf] = vini/vinf
    return umin,umax,cone
# resonant or fluorescent emission
def resfluor(w,u,vinf,beta,vini,geo,cone,par,VF,DP):
    omega  = vinf*w                                    # relative velocity
    Lrnzt1 = sqrt(1-omega**2)                          # Lorenzt factor
    DopRel = square(1-omega)*sqrt((1+omega)/(1-omega)) # relative Doppler shift
    dTheta = Lrnzt1/(omega*DopRel)                     # total polar term
    # azimuthal geometry
    # sphere
    if geo[1] >= 1-2**-20 and geo[8] <= 2**-20 :
        ell = 1
    # hemisphere
    elif geo[1] >= 1-2**-20 and geo[8] >= 1-2**-20 :
        ell = array(list(map(partial(hemi_inscr,u,vinf,*geo[6:11],cone),w)))
    # cones
    elif geo[1] < 1-2**-20 and geo[8] < 2**-20 :
        ell  = array(list(map(partial(cone_inscr,u,vinf,*geo[:11],cone),w)))
        # cavity in cone
        if geo[12] > 0. :
            ell -= array(list(map(partial(cone_inscr,u,vinf,*geo[11:],cone),w)))
    # cones with disk
    else:
        ell  = array(list(map(partial(cndk_inscr,u,vinf,*geo[:11],cone),w)))
        # cavity in cone
        if geo[12] > 0. :
            ell -= array(list(map(partial(cndk_inscr,u,vinf,*geo[11:],cone),w)))
        ell = nanmax([ell,zeros(len(w))],axis=0)
    return ell*n[DP](w,beta,vini,VF,*par)*dxdw(w,beta,vini,VF)* dTheta
# integral over column densities for range of allowed velocities
def phi_int(vinf,beta,incl,tO,tC,vdisk,vini,par,VF,DP,u,umin,umax,cone):
    # if cone projection excludes some velocities, then limit the integral
    if incl+tO < pi/2 :
        umax = min([umax,u/cos(incl+tO)])
    if incl > tO :
        umin   = max([umin,u/cos(incl-tO)])
    # if minimum exceeds maximum due to geometry limits,
    # no velocity bands are contained
    if umin >= umax:
        return 0
    else:
        # set up geometry terms that do not depend on velocity
        geo = precalc_geometry(incl,tO,tC,vdisk)
        # return the integral
        return fixed_quad(resfluor,umin,umax,args=(u,vinf,beta,vini,geo,cone,par,VF,DP),n=64)[0]
# calculate unnormalized profile for a sphere or bicone
def calc_phi(ww0,vinf,beta,incl,tO,tC,xdisk,vini,vapr,*par,VF='BetaCAK',DP='PowerLaw',Pulse='Normal'):
    # obtain velocity limits for integral
    umin,umax,cone = calc_limits(ww0,vinf,vini,vapr,beta,VF)
    # line of sight velocity with relativistic corrections
    u = absolute(((ww0)**2-1)/((ww0)**2+1)) / vinf
    # check disk radius and convert to velocity
    if xdisk <= 1 :
        vdisk = 2**-20
    elif xdisk == inf :
        vdisk = 1
    else:
        vdisk = max([v[VF](xdisk,beta,vini),2**-20])
    # profile integrated at each velocity
    phi = zeros(len(ww0))
    phi[ u < 1 ] = array( list( map( \
            partial(phi_int,vinf,beta,incl,tO,tC,vdisk,vini,par,VF,DP),\
            u[ u < 1 ], umin, umax, cone ) ) )
    return phi



from OutLines.__funcs__ import *
import matplotlib.pyplot as plt
from matplotlib.patches import Arc,Rectangle
from matplotlib import rcParams
from cycler import cycler
# vygrboi color scheme -- github.com/sflury/vygrboi
rcParams['axes.prop_cycle'] = cycler('color', ['#64008f','#fdca40','#08772f',\
                                    '#ab0014','#008cf7','#da6600','#001e52'])
rcParams['figure.figsize']       = (5,4.5) # figure size in inches
rcParams['xtick.top']            = True    # draw ticks on the top side
rcParams['xtick.bottom']         = True    # draw ticks on the bottom side
rcParams['xtick.major.size']     = 10      # major tick size in points
rcParams['xtick.minor.size']     = 4       # minor tick size in points
rcParams['xtick.major.width']    = 1       # major tick width in points
rcParams['xtick.direction']      = 'in'    # direction'] {in, out, inout}
rcParams['xtick.minor.visible']  = True    # visibility of minor ticks on x-axis
rcParams['ytick.left']           = True    # draw ticks on the left side
rcParams['ytick.right']          = True    # draw ticks on the right side
rcParams['ytick.major.size']     = 10      # major tick size in points
rcParams['ytick.minor.size']     = 4       # minor tick size in points
rcParams['ytick.major.width']    = 1       # major tick width in points
rcParams['ytick.direction']      = 'in'    # direction'] {in, out, inout}
rcParams['ytick.minor.visible']  = True    # visibility of minor ticks on y-axis


'''
Name:
    PlotIsoContours

Purpose:
    Plot isovelocity contours for a given velocity field and velocity index.

Keyword Arguments:
    :VelocityIndex (*float*): velocity field power index, default = 1
    :VelocityField (*str*): the desired velocity field, default = \'BetaCAK\'

Attributes:
    :fig (*maplotlib.pyplot.figure*): figure instance
    :ax  (*matplotlib.pyplot.axis*): axis instance
'''
class PlotIsoContours(object):
    def __init__(self,VelocityIndex=1,VelocityField='BetaCAK',V0=0,VT=1):
        fig,ax = self.isovel(beta=VelocityIndex,VelocityField=VelocityField,V0=V0)
        self.fig = fig
        self.ax  = ax
    def isovel(self,beta=1,VelocityField='BetaCAK',V0=0):
        rscl = 10
        figure,axis = plt.subplots(1,1)
        axis.annotate('',xy=(1.02,0),xytext=(1/rscl-0.01,0),arrowprops={'arrowstyle':'->'},zorder=4)

        axis.plot([0,0],[-1,-1/rscl],'-',lw=1,color='k')
        axis.plot([0,0],[ 1/rscl, 0.2],'-',lw=1,color='k')
        axis.plot([0,0],[ 0.900,  1.0],'-',lw=1,color='k')
        axis.text(-0.05,0.26,'plane of sky',rotation=270)
        axis.text(1.07,-0.1,'to\nobserver',horizontalalignment='center')

        theta = linspace(0,2*pi,1001)
        plt.plot(cos(theta)/rscl,sin(theta)/rscl,color='k')
        for u,color in zip([-0.5,-0.3,-0.1,0.1,0.3,0.5],['C3','C5','C1','C2','C4','C0']):
            if hasattr(beta,'__len__'):
                for b,ls in zip(beta,[':','-','-.','--']):
                    uw = linspace(abs(u),v[VelocityField](rscl*0.85,b,V0),1001)
                    rbnd = x[VelocityField](uw,b,V0)/rscl
                    tr = arccos(abs(u)/uw)
                    xt = rbnd*cos(tr)
                    yt = rbnd*sin(tr)
                    axis.plot(sign(u)*xt+u, yt,color=color,ls=ls,lw=2)
                    axis.plot(sign(u)*xt+u,-yt,color=color,ls=ls,lw=2)
            else:
                uw = linspace(abs(u),v[VelocityField](rscl*0.9,beta,V0),1001)
                rbnd = x[VelocityField](uw,beta,V0)/rscl
                tr = arccos(abs(u)/uw)
                xt = rbnd*cos(tr)
                yt = rbnd*sin(tr)
                axis.plot(sign(u)*xt+u, yt,color=color,ls='-',lw=2)
                axis.plot(sign(u)*xt+u,-yt,color=color,ls='-',lw=2)
        axis.set_aspect('equal')
        axis.axis('off')
        axis.text(-1,1,VelocityField,fontsize=14,fontweight='bold')
        axis.set_xlim(-1.1,1.2)
        axis.set_ylim(-1.1,1.2)
        plt.subplots_adjust(left=0,right=1,top=1,bottom=0)
        return figure,axis

'''
Name:
    PlotGeometry

Purpose:
    Plot an edge-on visualization of the geometry of a directed outflow with
    optional cavity and optional disk.

Arguments:
    :i (*float*): inclination of the outflow with respect to the line of
        sight; if i > 6.3, interpreted as units of degrees, otherwise as radians
    :t (*float*): opening angle of the outflow with respect to the inclination;
        if t > 6.3, interpreted as units of degrees, otherwise as radians

Keyword Arguments:
    :c (*float*):  cavity angle of the outflow with respect to the inclination;
        if t > 6.3, interpreted as units of degrees, otherwise as radians
    :xd (*float*): disk radius relative to the launch radius, must be > 1

Attributes:
    :fig (*maplotlib.pyplot.figure*): figure instance
    :ax  (*matplotlib.pyplot.axis*): axis instance
'''
class PlotGeometry(object):
    def __init__(self,i,t,c=0,xd=0):
        if i > pi/2 :
            i *= pi/180
        if t > pi/2 :
            t *= pi/180
        if c > pi/2 :
            c *= pi/180
        if c > t :
            print('Opening angle must be greater than cavity angle.')
        else:
            fig,ax = self.plot_geom(i,t,c,xd)
            self.fig = fig
            self.ax  = ax
    def plot_geom(self,i,t,c,xd):
        theta = linspace(0,2*pi,1001)
        theth = linspace(0,pi,1001)
        tcone = linspace(i-t,i+t,1001)
        figure,axis = plt.subplots(1,1,figsize=(4.5,4.5),sharex=True,sharey=True)
        axis.set_aspect('equal')
        axis.axis('off')
        axis.set_xlim(-5.1,5.1)
        axis.set_ylim(-5.1,5.1)
        axis.plot([0,0],[-20,20],'--',color='k')
        axis.text(-0.55,4,'plane\nof sky',rotation=270)
        axis.annotate('',xy=(5.1,0),xytext=(-0.2,0),arrowprops={'arrowstyle':'->'},zorder=4)
        axis.text(2,-0.5,'to observer')
        # plot outflow cone
        axis.plot([5*cos(i-t+pi),5*cos(i-t)],[5*sin(i-t+pi),5*sin(i-t)],color='k',lw=2,zorder=4)
        axis.plot([5*cos(i+t+pi),5*cos(i+t)],[5*sin(i+t+pi),5*sin(i+t)],color='k',lw=2,zorder=4)

        axis.plot([0,5*cos(i)],[0,5*sin(i)],'--',color='k') # bisecting inclination angle
        axis.add_patch(Arc((0,0),3.6,3.6,angle=0,theta1=0,theta2=i*180/pi,color='C3',lw=2))
        axis.text(2.1*cos(i/2),1.9*sin(i/2),r'$i$',color='C3',horizontalalignment='center')
        axis.add_patch(Arc((0,0),5.8,5.8,angle=0,theta1=(i-t)*180/pi,theta2=i*180/pi,color='C2',lw=1.5))
        axis.add_patch(Arc((0,0),6.2,6.2,angle=0,theta1=(i-t)*180/pi,theta2=i*180/pi,color='C2',lw=1.5))
        axis.text(3.1*cos(i-t/2),4*sin(i-t/2),r'$\theta_o$',color='C2')
        # plot hollow inner part of cone
        if c > 0 :
            axis.plot([5*cos(i-c+pi),5*cos(i-c)],[5*sin(i-c+pi),5*sin(i-c)],color='k',lw=2,zorder=4)
            axis.plot([5*cos(i+c+pi),5*cos(i+c)],[5*sin(i+c+pi),5*sin(i+c)],color='k',lw=2,zorder=4)
            axis.add_patch(Arc((0,0),4.1,4.1,angle=0,theta1=i*180/pi,theta2=(i+c)*180/pi,color='C5',lw=1.5))
            axis.add_patch(Arc((0,0),4.3,4.3,angle=0,theta1=i*180/pi,theta2=(i+c)*180/pi,color='C5',lw=1.5))
            axis.add_patch(Arc((0,0),4.5,4.5,angle=0,theta1=i*180/pi,theta2=(i+c)*180/pi,color='C5',lw=1.5))
            axis.text(2.3*cos(i+c/2),2.4*sin(i+c/2),r'$\theta_c$',color='C5')
        # plot disk
        if xd > 1 :
            axis.fill_between([-10,-sin(i)*xd,sin(i)*xd],[-cos(i)*xd,-cos(i)*xd,-cos(i)*xd],[cos(i)*xd,cos(i)*xd,-cos(i)*xd],color='0.5',alpha=0.5,edgecolor='none')
            dr = 0.1
            axis.add_patch(Rectangle((-xd,-dr),2*xd,2*dr,angle=(i-pi/2)*180/pi,rotation_point=(0,0),facecolor='w',edgecolor='k',zorder=10))
            axis.text(-0.8*sin(i)*xd,0.6*cos(i)*xd,'disk',rotation=(i-pi/2)*180/pi)
        # plot source
        axis.fill_between([-10,0],[-1,-1],[1,1],color='0.5',alpha=0.5,edgecolor='none')
        axis.fill_between(cos(theth),sin(theth+pi)[::-1],sin(theth),color='w',edgecolor='k',zorder=10)
        axis.text(-0.8,-0.15,'source',zorder=11)
        plt.subplots_adjust(left=0,right=1,top=0.95,bottom=0.05)
        return figure,axis
'''
Name:
    PlotConeProjection

Purpose:
    Plot an edge-on and face-on visualization of the directed outflow cone
    projections onto the plane of the sky, including three representative
    band contours illustrating the fraction of a given de-projected velocity
    inscribed by the projected outflow.

Arguments:
    :i (*float*): inclination of the outflow with respect to the line of
        sight; if i > 6.3, interpreted as units of degrees, otherwise as radians
    :t (*float*): opening angle of the outflow with respect to the inclination;
        if t > 6.3, interpreted as units of degrees, otherwise as radians
    :c (*float*): cavity angle of the outflow with respect to the inclination;
        if t > 6.3, interpreted as units of degrees, otherwise as radians

Attributes:
    :fig (*maplotlib.pyplot.figure*): figure instance
    :ax  (*matplotlib.pyplot.axis*): axis instance
'''
class PlotConeProjection(object):
    def __init__(self,i,t,c=0):
        if i > pi/2 :
            i *= pi/180
        if t > pi/2 :
            t *= pi/180
        if c > pi/2 :
            c *= pi/180
        if c > t :
            print('Opening angle must be greater than cavity angle.')
        else:
            fig,ax = self.plot_geom(i,t,c)
            self.fig = fig
            self.ax  = ax
    @staticmethod
    def plot_cone(axis,i,t,color='C5'):
        theta = linspace(0,2*pi,1001)
        theth = linspace(0,pi,1001)
        tcone = linspace(i-t,i+t,1001)
        axis.plot(cos(theta),sin(theta),color='k',lw=2.5)
        axis.annotate('',xy=(1.02,0),xytext=(-0.02,0),arrowprops={'arrowstyle':'->'},zorder=4)
        axis.plot(cos(tcone),sin(tcone),lw=2.5,color=color,zorder=5)
        axis.plot(cos(tcone+pi),sin(tcone+pi),lw=2.5,color=color,zorder=5)
        axis.plot([cos(i-t+pi),cos(i-t)],[sin(i-t+pi),sin(i-t)],color='k',lw=1,zorder=4)
        axis.plot([cos(i+t+pi),cos(i+t)],[sin(i+t+pi),sin(i+t)],color='k',lw=1,zorder=4)
        #
        xi = cos(i)*cos(t)
        yi = sin(i)*cos(t)
        xp  = sin(t)
        yp  = 0.1*cos(i)*xp
        A  = i - 3*pi/2
        xt1 = xp*cos(theth)*cos(A)-yp*sin(theth)*sin(A)
        yt1 = xp*cos(theth)*sin(A)+yp*sin(theth)*cos(A)
        xt2 = xp*cos(theth+pi)*cos(A)-yp*sin(theth+pi)*sin(A)
        yt2 = xp*cos(theth+pi)*sin(A)+yp*sin(theth+pi)*cos(A)
        axis.plot(xt1+xi,yt1+yi,color='C5',lw=2.5,zorder=5)
        axis.plot(xt1-xi,yt1-yi,color='C5',lw=2.5,zorder=5)
        axis.plot(xt2+xi,yt2+yi,color='C5',lw=2.5,ls=':',zorder=3)
        axis.plot(xt2-xi,yt2-yi,color='C5',lw=2.5,ls=':',zorder=3)
        return axis
    @staticmethod
    def plot_band(axis,i,t,u,w,c=0,color='C0'):
        # angle of velocity projection
        rbnd = sqrt(1-(u/w)**2)
        # outer cone projected as ellipse
        f0 = sin(i)*cos(t)
        g0 = sin(t)
        h0 = g0*cos(i)
        # intersection points of band contour with outer cone projections
        a0 = (g0/h0)**2-1
        b0 = -2*f0*(g0/h0)**2
        c0 = rbnd**2+((f0/h0)**2-1)*g0**2
        q0 = (-b0+array([-1,1])*sqrt(b0**2-4*a0*c0))/(2*a0)
        # inner cone projected as ellipse
        f1 = sin(i)*cos(c)
        g1 = sin(c)
        h1 = g1*cos(i)
        # intersection points of band contour with inner cone projections
        a1 = (g1/h1)**2-1
        b1 = -2*f1*(g1/h1)**2
        c1 = rbnd**2+((f1/h1)**2-1)*g1**2
        q1 = (-b1+array([-1,1])*sqrt(b1**2-4*a1*c1))/(2*a1)
        if c == 0:
            q1 = [0,0]

        theta = linspace(-pi/2,3*pi/2,1001)
        xt = rbnd*cos(theta)
        yt = rbnd*sin(theta)
        if i == pi/2 :
            ind = where(((yt>-q0[0])&(yt<q0[0]))|(yt<-q1[0])|(yt>q1[0]))[0]
        elif i + t > pi/2 :
            ind = where(((yt>q1[0])&(yt<q1[1]))|((yt>-q0[1])&(yt<q0[0])))[0]
        elif i == 0 :
            if rbnd < h0-f0 :
                ind = array([False for j in range(len(yt))])
            else:
                ind = array([True  for j in range(len(yt))])
        else:
            ind = where(((yt>q1[0])&(yt<q1[1]))|(yt>q0[1])|(yt<q0[0]))[0]
        xt[ind] = nan
        yt[ind] = nan
        axis.plot(0.1*xt[501:]+u/w,yt[501:],zorder=4,color=color)
        axis.plot(0.1*xt[:501]+u/w,yt[:501],zorder=4,color=color,ls=':')
        xt[where(isfinite(xt))[0]] = nan
        yt[where(isfinite(yt))[0]] = nan
        xt[ind] = rbnd*cos(theta[ind])
        yt[ind] = rbnd*sin(theta[ind])

        axis.plot(0.1*xt[501:]+u/w,yt[501:],zorder=4,color=color,alpha=0.2)
        axis.plot(0.1*xt[:501]+u/w,yt[:501],zorder=4,color=color,alpha=0.2,ls=':')

        return axis
    @staticmethod
    def proj_cone(axis,i,t,color='C5'):
        theta = linspace(0,2*pi,1001)
        axis.plot(cos(theta),sin(theta),'-k',lw=2.5,zorder=0)
        if t >= pi/2 :
            axis.plot(cos(theta),sin(theta),'C5',lw=2.5)
            return axis
        # cone projected as ellipse
        f = sin(i)*cos(t)
        g = sin(t)
        h = g*cos(i)
        # intersection with unit sphere
        a = (g/h)**2-1
        b = -2*f*(g/h)**2
        xc = 1+((f/h)**2-1)*g**2
        xq = (-b+array([-1,1])*sqrt(b**2-4*a*xc))/(2*a)
        xp = sqrt(1-xq**2)
        de = arccos(xq[0])
        # flat lines at i = 90 degrees
        if i == pi/2:
            axis.plot([-g,g],[h+f,h+f],lw=2.5,color='C5')
            axis.plot([-g,g],[h-f,h-f],lw=2.5,color='C5')
            tc = linspace(pi/2-de,pi/2+de,1001)
            axis.plot(cos(tc),sin(tc),lw=2.5,color=color)
            tc = linspace(3*pi/2-de,3*pi/2+de,1001)
            axis.plot(cos(tc),sin(tc),lw=2.5,color=color)
        # if intersection with unit circle, account for lune
        # from cap projection close to transverse
        elif i + t > pi/2 :
            if isnan(xq[0]) :
                xe = g*cos(theta)
                ye = h*sin(theta)+f
                xq = 2*[ye[argmax((xe**2+ye**2))]]
                de = arccos(xq[0])
            # ellipse + lune
            xe = g*cos(theta)
            ye = h*sin(theta)+f
            xe[ye>xq[0]] = nan
            ye[ye>xq[0]] = nan
            tc = linspace(pi/2-de,pi/2+de,1001)
            axis.plot(cos(tc),sin(tc),lw=2.5,color=color)
            axis.plot(xe,ye,color='C5',lw=2.5)
            # lune
            xe = g*cos(theta)
            ye = h*sin(theta)-f
            xe[ye>-xq[1]] = nan
            ye[ye>-xq[1]] = nan
            tc = linspace(3*pi/2-de,3*pi/2+de,1001)
            axis.plot(cos(tc),sin(tc),lw=2.5,color=color)
            axis.plot(xe,ye,color='C5',lw=2.5)
        # if no intersection, just an ellipse
        else:
            xe = g*cos(theta)
            ye = h*sin(theta)+f
            axis.plot(xe,ye,color='C5',lw=2.5)
        return axis
    @staticmethod
    def proj_band(axis,i,t,u,w,c=0,color='C0'):
        # angle of velocity projection
        rbnd = sqrt(1-(u/w)**2)
        # outer cone projected as ellipse
        f0 = sin(i)*cos(t)
        g0 = sin(t)
        h0 = g0*cos(i)
        # intersection points of band contour with outer cone projections
        a0 = (g0/h0)**2-1
        b0 = -2*f0*(g0/h0)**2
        c0 = rbnd**2+((f0/h0)**2-1)*g0**2
        q0 = (-b0+array([-1,1])*sqrt(b0**2-4*a0*c0))/(2*a0)
        # inner cone projected as ellipse
        f1 = sin(i)*cos(c)
        g1 = sin(c)
        h1 = g1*cos(i)
        # intersection points of band contour with inner cone projections
        a1 = (g1/h1)**2-1
        b1 = -2*f1*(g1/h1)**2
        c1 = rbnd**2+((f1/h1)**2-1)*g1**2
        q1 = (-b1+array([-1,1])*sqrt(b1**2-4*a1*c1))/(2*a1)
        if c == 0:
            q1 = [0,0]

        theta = linspace(-pi/2,3*pi/2,1001)
        xt = rbnd*cos(theta)
        yt = rbnd*sin(theta)
        axis.plot(xt,yt,color=color,alpha=0.2)
        if i == pi/2 :
            ind = where(((yt>-q1[0])&(yt<q1[0]))|(yt<-q0[0])|(yt>q0[0]))[0]
        elif i + t > pi/2 :
            if i + c > pi/2 : #q1[1] >= 1 :
                ind = where((yt<-q1[1])|(yt>q1[0])|((yt>-q0[1])&(yt<q0[0])))[0]
            else:
                ind = where(((yt>q1[0])&(yt<q1[1]))|((yt>-q0[1])&(yt<q0[0])))[0]
        elif q0[0] < 0 :
            ind = where(((yt>q1[0])&(yt<q1[1]))|(yt<q0[0]))[0]
        elif i == 0 :
            if rbnd < h0-f0 :
                ind = array([False for j in range(len(yt))])
            else:
                ind = array([True  for j in range(len(yt))])
        else:
            ind = where(((yt>q1[0])&(yt<q1[1]))|(yt>q0[1])|(yt<q0[0]))[0]
        xt[ind] = nan
        yt[ind] = nan
        axis.plot(xt,yt,color=color,zorder=0)

        return axis
    def plot_geom(self,i,t,c=0):
        # convert to degrees
        istr = f'{int(round(i*180/pi)):0>2d}'
        tstr = f'{int(round(t*180/pi)):0>2d}'
        # plotting
        fig,ax = plt.subplots(1,2,figsize=(7,4),sharex=True,sharey=True)
        #plot the cones
        ax[0] = self.plot_cone(ax[0],i,t,color='C5')
        ax[1] = self.proj_cone(ax[1],i,t,color='C5')
        if c > 0 :
            ax[0] = self.plot_cone(ax[0],i,c,color='k')
            ax[1] = self.proj_cone(ax[1],i,c,color='k')
        ax[0].set_title('Edge-On View')
        ax[1].set_title('Plane of Sky')
        for uj,cj in zip([0.3,0.45,0.6],['C2','C4','C0']):
            # band contours
            ax[0] = self.plot_band(ax[0],i,t,uj,0.75,color=cj,c=c)
            ax[1] = self.proj_band(ax[1],i,t,uj,0.75,color=cj,c=c)
        # label
        fig.text(0.45,0.9,r'$i\ =$'+f'{int(round(i*180/pi)): >3d}'+r'$^\circ$')
        fig.text(0.45,0.825,r'$\theta_o\ =$'+f'{int(round(t*180/pi)): >3d}'+r'$^\circ$')
        if c != 0 :
            fig.text(0.45,0.75,r'$\theta_c\ =$'+f'{int(round(c*180/pi)): >3d}'+r'$^\circ$')
        #ax[1].text(0,0,r'$\oplus$',color='k')
        for a in ax:
            a.set_aspect('equal')
            a.axis('off')
            a.set_xlim(-1.1,1.1)
            a.set_ylim(-1.1,1.1)
        plt.subplots_adjust(left=0,right=1,top=0.95,bottom=0)
        return fig,ax



'''
name:
    OutLines
dependencies:
    python     3.12.9
    numpy      2.3.5
    scipy      1.14.1
    matplotlib 3.10.0
author:
    Sophia Flury 2026.02.09
'''
# OutLines functions
import OutLines.__absorption__
import OutLines.__nebular__
import OutLines.__resonfluor__
import OutLines.__funcs__
import OutLines.__params__
import OutLines.__properties__
import OutLines.__profiles__
import OutLines.__visualize__
# add class calls directly to OutLines for convenience
OutLines.Nebular             = OutLines.__profiles__.Nebular
OutLines.Absorption          = OutLines.__profiles__.Absorption
OutLines.Resonant            = OutLines.__profiles__.Resonant
OutLines.Fluorescent         = OutLines.__profiles__.Fluorescent
OutLines.Properties          = OutLines.__properties__.Properties
OutLines.PlotGeometry        = OutLines.__visualize__.PlotGeometry
OutLines.PlotConeProjection  = OutLines.__visualize__.PlotConeProjection
OutLines.PlotIsoContours     = OutLines.__visualize__.PlotIsoContours
# quote from Shelly's "Ode to the West Wind"
def Quote():
    print('\nWild Spirit, which art moving everywhere;\n'+\
          'Destroyer and preserver\n'+
          'on whose stream, mid the steep sky\'s commotion,\n'+
          'Loose clouds Shook from the tangled boughs of Heaven.\n'+\
          12*' '+' ~ Percy Bysshe Shelley, \"Ode to the West Wind\"')



from OutLines import __nebular__,__absorption__,__resonfluor__
from OutLines.__funcs__ import *
from OutLines.__params__ import *
# numpy functions for some basics
from numpy.random import randn
import sys
# wrapper to apply meta Profile object class to specific line profile cases
# with generalized methods and attributes
def profile_constructor(ProfileSubClass):
    '''
    Name:
        profile_constructor

    Purpose:
        Construct the Profile class given a particular line profile object

    Arguments:
        :ProfileSubClass (*class*): object class of `Nebular`, `Resonant`,
                `Fluorescent` or `Absorption`

    Returns:
        :Profile (*object*): An instance of the Profile object class inhereting
                the Nebular, Resonant, Fluorescent, or Absorption class with
                properties specific to the related spectral line

    '''
    class Profile(ProfileSubClass):
        '''
        Name:
            Profile

        Purpose:
            Generate a model line profile for the given outflow conditions
            at the given wavelength (and oscilator strength if applicable)

        Keyword Arguments:
            :VelocityField (*str*): radial velocity field indicating the assumed
                    treatment of the underlying physics and acceleration.
                    Options include
                         \'BetaCAK\' (1-1/x)^b from CAK theory - Castor et al. 1975
                         \'AccPlaw\' sqrt(1-x^(1-b)) from Steidel et al. 2010
                         \'VelPlaw\' 0.5(x-1)^b from various sources
            :Geometry (*str*): azimuthal geometry capturing the (an)isotropy
                    Options include
                        \'Sphere\' or \'Spherical\' isotropic case
                        \'FilledCones\' bidirectional cones
                        \'HollowCones\' or \'OpenCones\' cones with a cavity
                        \'HollowConesFixedCavity\' cones with fixed cavity size
            :DensityProfile (*str*): radial profile to use for the gas density.
                    Options include integration of many pulses or episodes
                         \'PowerLaw\'
                         \'Exponential\'
                         \'PowerLaw2\'
                     and individual pulse-like episodes or outbursts
                         \'LogNormal\'
                         \'Normal\'
                         \'Shell\'
                         \'FRED\'
                    or a combination of the two
                         \'Pulses\'
                         \'DampedPulses\'
            :Disk     (*bool*): boolean indicating whether a disk is present and
                    obstructs the cone posterior to the observer -- requires cone
                    geometry in order to take effect, default is False
            :FromRest (*bool*): boolean indicating whether the wind/shell/bubble
                    is launched from rest, default is True
            :Aperture (*bool*): boolean indicating whether the observed profile is
                    truncated due to the effects of a spherical aperture, default
                    is False

        Attributes:
            :Profile (*function*): function returning a line profile based on the chosen
                    velocity field, geometry, and density profile; takes wavelength and
                    paramaters (see the `params` attribute)
            :OutLinesModel (*function*): function returning the normalized
                    OutLines profile for a wind, bubble, or outflow without
                    fluxes, column densities, or other radiative transfer terms
            :npar (*int*): number of assigned parameters
            :params (*dict*): dictionary of parameters, initially assigned or later
                    updated by the user
            :bounds (*dict*): dictionary of parameter bounds, initially assigned or
                    later updated by the user

        Bound methods for viewing, updating, and computing the line profile:
            :get_profile: returns a model line profile given a wavelength array
            :get_params: returns the current parameters as a list
            :get_param_names: returns parameter names as a list
            :get_bounds: returns the current boundaries set for the parameters
            :update_params: updates parameter(s) given the name(s) and value(s)
            :update_bounds: updates bound(s) given the name(s) and value(s)
            :print_params: prints current parameter values to the command line
            :print_bounds: prints current parameter bounds to the command line
            :print_settings: prints profile model settings to the command line
            :set_params: resets parameter values to the OutLines defaults
            :set_bounds: resets paramater bounds to the OutLines defaults

        Bound methods for obtaining velocity quantiles of the line profile:
            :velocity_quantiles: returns velocity quantiles
            :W80: returns the W80 (0.9-0.1) velocity quantile
            :FWHM: returns the 0.75-0.25 velocity quantile

        Bound methods for modeling an observed line profile with OutLines:
            :init_params: samples 2*npar+1 values of the current parameters
                    assuming some intrinsic dispersion (default of 2^-6);
                    designed to generate valid initial walkers for MCMC fits
            :log_prior: returns a uniform bounded prior on the parameters with
                    parameters bounds defined in the Profile attribute
            :log_likelihood: returns the chi squared log likelihood of the line
                    profile model generated by OutLines given wavelength,
                    flux, and flux uncertainties
            :log_probability: returns the log probability of the parameters
                    given the data; sums the log prior and the log likelihood;
                    can be passed to MCMC codes like \'emcee\'
        '''
        def __init__(self,*args,VelocityField='BetaCAK',DensityProfile='PowerLaw',Pulse='Normal',\
                    Geometry='Spherical',AddStatic=False,Disk=False,Aperture=False,FromRest=True):

            # keyword arguments specifying the model
            self.settings = {'Profile':ProfileSubClass.__name__,\
                             'VelocityField':VelocityField,\
                             'DensityProfile':DensityProfile,\
                             'Geometry':Geometry.replace('Open','Hollow'),\
                             'StaticComponent':AddStatic,\
                             'FromRest':FromRest,\
                             'Aperture':Aperture,\
                             'Disk':Disk
                             }
            if 'Pulse' in self.settings['DensityProfile'] :
                self.settings['Pulse'] = Pulse
            # inheret the wind
            super(Profile,self).__init__(*args)

            # set central wavelength(s)
            if hasattr(args[0],'__len__'):
                self.w0 = array(args[0])
                # if absorption is present, set the oscilator strength if given
                if ProfileSubClass.__name__ != 'Nebular':
                    if len(args[0]) == len(args[1]):
                        self.fosc = array(args[1])
                    # otherwise, inform the user and call it quits
                    else:
                        print('OutLines requires the same number of oscilator'+\
                            '\nstrengths as central wavelengths for '+\
                            '\nAbsorption, Resonant, and Fluorescent line profiles')
                        sys.exit()
                    if ProfileSubClass.__name__ != 'Absorption':
                        if ProfileSubClass.__name__ != 'Fluorescent' :
                            if len(args[1]) == len(args[2]):
                                self.pline = array(args[2])
                            else:
                                print('OutLines requires the same number of oscilator'+\
                                    '\nstrengths and channel emission fractions for '+\
                                    '\nResonant and P Cygni line profiles')
                                sys.exit()
                        if ProfileSubClass.__name__ == 'Fluorescent' :
                            if len(args[1]) == len(args[2]) and len(args[1]) == len(args[3]):
                                self.fres  = array(args[2])
                                self.pline = array(args[3])
                            else:
                                print('OutLines requires the same number of oscilator'+\
                                    '\nstrengths and channel emission fractions for '+\
                                    '\nFluorescent line profiles')
                                sys.exit()
            else:
                self.w0 = array([args[0]])
                # if absorption is present, set the oscilator strength if given
                if ProfileSubClass.__name__ != 'Nebular':
                    self.fosc = array([args[1]])
                    if ProfileSubClass.__name__ != 'Absorption':
                        if ProfileSubClass.__name__ != 'Fluorescent' :
                            self.pline = array([args[2]])
                        else:
                            self.fres  = array([args[2]])
                            self.pline = array([args[3]])
            # set the number of lines
            self.nLines = len(self.w0)

            # set parameters, bounds, and profile model
            self.set_params()
            self.set_profile()
        # print documentation
        def docs(self):
            print(self.__doc__)
        # setters for the profile function, parameters, and boundaries
        def set_profile(self):
            if self.settings['StaticComponent']:
                self.Profile = self.__WithStatic__
                self.get_outflow    = self.__get__outflow_profile__
                self.get_static     = self.__get__static_profile__
            else:
                self.Profile = self.__WithoutStatic__
        def set_params(self):
            # set parameters, names, and labels
            # starting with velocities and fluxes
            if self.settings['StaticComponent']:
                par_init = [3e-5,1e-3,*BetaPars[self.settings['VelocityField']]]
                par_init += [ \
                    *ProfPars['Static'][self.settings['Profile']],\
                    *ProfPars['Outflow'][self.settings['Profile']] ]
                par_name  = ['DopplerWidth','TerminalVelocity','VelocityIndex']
                par_name += [ \
                    *ProfName['Static'][self.settings['Profile']],\
                    *ProfName['Outflow'][self.settings['Profile']]]
                par_labs  = [r'$\sigma_v$ [km s$^{-1}$]',r'$v_\infty$ [km s$^{-1}$]',r'$\beta$']
                par_labs += [ \
                    *ProfLabs['Static'][self.settings['Profile']],\
                    *ProfLabs['Outflow'][self.settings['Profile']] ]
            else:
                par_init = [1e-3,*BetaPars[self.settings['VelocityField']]]
                par_init += [ \
                    *ProfPars['Outflow'][self.settings['Profile']] ]
                par_name  = ['TerminalVelocity','VelocityIndex']
                par_name += [ \
                    *ProfName['Outflow'][self.settings['Profile']] ]
                par_labs  = [r'$v_\infty$ [km s$^{-1}$]',r'$\beta$']
                par_labs += [ \
                    *ProfLabs['Outflow'][self.settings['Profile']] ]
            # add in geometry
            par_init += [*GeomPars[self.settings['Geometry']]]
            par_name += [*GeomName[self.settings['Geometry']]]
            par_labs += [*GeomLabs[self.settings['Geometry']]]
            # optional disk
            if self.settings['Disk'] :
                par_init += [2]
                par_name += ['DiskRadius']
                par_labs += [r'$x_{disk}$']
            # optional launch velocity
            if not self.settings['FromRest'] :
                par_init += [1e-6]
                par_name += ['LaunchVelocity']
                par_labs += [r'$v_{0}$']
            # optional aperture limit
            if self.settings['Aperture'] :
                par_init += [1e-3]
                par_name += ['ApertureVelocity']
                par_labs += [r'$v_{apert}$']
            par_init += [*DensPars[self.settings['DensityProfile']]]
            par_name += [*DensName[self.settings['DensityProfile']]]
            par_labs += [*DensLabs[self.settings['DensityProfile']]]
            # sort and store parameters in dictionaries
            self.params = {}
            self.labels = {}
            if self.settings['StaticComponent'] :
                self.statps = {}
                self.outfps = {}
            for name,val,lab in zip(par_name,par_init,par_labs):
                if 'Flux' in name or 'Column' in name :
                    for i in range(self.nLines):
                        self.params[f'{name}{i+1}'] = val
                        self.labels[f'{name}{i+1}'] = lab+f'$_{i+1}$'
                else:
                    self.params[name] = val
                    self.labels[name] = lab
                if self.settings['StaticComponent'] :
                    if 'Static' in name or 'Doppler' in name:
                        if 'Flux' in name or 'Column' in name:
                            for i in range(self.nLines):
                                self.statps[f'{name}{i+1}'] = val
                        else:
                            self.statps[name] = val
                    else:
                        if 'Flux' in name or 'Column' in name :
                            for i in range(self.nLines):
                                self.outfps[f'{name}{i+1}'] = val
                        else:
                            self.outfps[name] = val
            self.npar = len(self.params)
            self.gpname = GeomName[self.settings['Geometry']]
            self.npname = DensName[self.settings['DensityProfile']]
            # set parameter bounds
            bounds = [[],[]]
            for l in [0,1]:
                if self.settings['StaticComponent']:
                    bounds[l] += [float(l),float(l),\
                        *BetaBounds[self.settings['VelocityField']][l],\
                        *ProfBounds['Static'][self.settings['Profile']][l],\
                        *ProfBounds['Outflow'][self.settings['Profile']][l] ]
                else:
                    bounds[l] += [float(l),\
                        *BetaBounds[self.settings['VelocityField']][l],\
                        *ProfBounds['Outflow'][self.settings['Profile']][l] ]
                bounds[l] += [*GeomBounds[self.settings['Geometry']][l]]
                if self.settings['Disk'] :
                    bounds[l] += [*DiskBounds[l]]
                if not self.settings['FromRest'] :
                    bounds[l] += [float(l)]
                if self.settings['Aperture'] :
                    bounds[l] += [float(l)]
                bounds[l] += [*DensBounds[self.settings['DensityProfile']][l]]
            # sort and store parameter bounds in dictionary
            self.bounds = {}
            for name,lower,upper in zip(par_name,bounds[0],bounds[1]):
                if 'Flux' in name or 'Column' in name :
                    for i in range(self.nLines):
                        name.strip(f'{i}')
                        self.bounds[f'{name}{i+1}'] = [lower,upper]
                else:
                    self.bounds[name] = [lower,upper]
        # getters for the names, labels, paramaters, bounds, and profile
        def get_params(self,comp='all'):
            if comp.lower() == 'outflow' and self.settings['StaticComponent']:
                return array(list(self.outfps.values()))
            elif comp.lower() == 'static' and self.settings['StaticComponent']:
                return array(list(self.statps.values()))
            else:
                return array(list(self.params.values()))
        def get_param_names(self):
            return array(list(self.params.keys()))
        def get_param_labels(self):
            return array(list(self.labels.values()))
        def get_bounds(self,astype=tuple):
            return astype(array(list(self.bounds.values())).T)
        def get_profile(self,wavelengths):
            return self.Profile(wavelengths,*self.get_params())
        def __get__static_profile__(self,wavelengths):
            return self.__OnlyStatic__(wavelengths,*self.get_params(comp='static'))
        def __get__outflow_profile__(self,wavelengths):
            return self.__WithoutStatic__(wavelengths,*self.get_params(comp='outflow'))
        # options for updating the parameters and bounds
        def update_params(self,par_name,par_val):
            if hasattr(par_val,'__len__'):
                for name,val in zip(par_name,par_val):
                    self.update_params(name,val)
            else:
                # if velocity > 1, convert to c units
                if 'Terminal' in par_name or 'Launch' in par_name or 'Doppler' in par_name or 'Aperture' in par_name :
                    if par_val > 1 :
                        par_val = par_val/2.99792458e5
                if 'Inclination' in par_name or 'Angle' in par_name:
                    if par_val > pi/2 :
                        if par_val > 90 :
                            print('Angles must be between 0 and 90 degrees')
                        par_val = par_val*pi/180
                self.params[par_name] = par_val
                if self.settings['StaticComponent'] :
                    if 'Static' in par_name or 'Doppler' in par_name:
                        self.statps[par_name] = par_val
                    else:
                        self.outfps[par_name] = par_val
        # options for updating the parameters and bounds
        def update_bounds(self,par_name,bound_val):
            if hasattr(bound_val[0],'__len__'):
                for name,val in zip(par_name,bound_val):
                    self.update_bounds(name,val)
            else:
                # if velocity > 1, convert to c units
                if 'Terminal' in par_name or 'Launch' in par_name or 'Doppler' in par_name or 'Aperture' in par_name :
                    if bound_val[1] > 1 :
                        bound_val[0] = bound_val[0]/2.99792458e5
                        bound_val[1] = bound_val[1]/2.99792458e5
                if 'Inclination' in par_name or 'Angle' in par_name:
                    if bound_val[1] > pi/2 :
                        if bound_val[1] > 90 :
                            print('Angles must be between 0 and 90 degrees')
                            bound_val[0] = bound_val[0]%90
                            bound_val[1] = bound_val[1]%90
                        bound_val[0] = bound_val[0]*pi/180
                        bound_val[1] = bound_val[1]*pi/180

                self.bounds[par_name] = bound_val
        # convenience function for printing settings to terminal
        def print_settings(self):
            print(3*' '+36*'-'+'\n  |'+11*' '+'MODEL SETTINGS'+11*' '+'|\n'+3*' '+36*'-')
            for i,w0 in enumerate(self.w0):
                print(f'  |     Line  {i+1: >2d}     : {w0: <15.3f} |')
            for option,setting in iter(self.settings.items()):
                if type(setting) == bool :
                    if setting:
                        print(f'  | {option: >16s} : {"Yes": <15s} |')
                    else:
                        print(f'  | {option: >16s} : {"No": <15s} |')
                elif 'Fixed' in setting :
                    s1 = setting.split('Fixed')[0]
                    s2 = 'Fixed'+setting.split('Fixed')[1]
                    print(f'  | {option: >16s} : {s1: <15s} |')
                    print(f'  | {"": >16s}   {s2: <15s} |')
                else:
                    print(f'  | {option: >16s} : {setting: <15s} |')
            print(3*' '+36*'-')
        # convenience function for printing parameters to terminal
        def print_params(self):
            print(3*' '+38*'-'+'\n  |'+11*' '+'MODEL PARAMETERS'+11*' '+'|\n'+3*' '+38*'-')
            for param,value in iter(self.params.items()):
                if 'Terminal' in param or 'Launch' in param or 'Doppler' in param or 'Aperture' in param :
                    print(f'  | {param: >18s} : {value*2.99792458e5: >10.3f} km/s |')
                elif 'Inclination' in param or 'Angle' in param:
                    print(f'  | {param: >18s} : {value*180/pi: >10.3f}°     |')
                else:
                    print(f'  | {param: >18s} : {value: >10.3f}      |')
            print(3*' '+38*'-')
        # convenience function for printing parameter bounds to terminal
        def print_bounds(self):
            delim = [',','|']
            print(3*' '+61*'-'+'\n  |'+20*' '+'PARAMETER  BOUNDARIES'+20*' '+'|\n'+3*' '+61*'-')
            for param,bound in iter(self.bounds.items()):
                line = f'  | {param: >18s} : '
                for l in [0,1]:
                    if 'Terminal' in param or 'Launch' in param or 'Doppler' in param or 'Aperture' in param :
                        line += f'{bound[l]*2.99792458e5: >13.3f} km/s {delim[l]}'
                    elif 'Inclination' in param or 'Angle' in param:
                        line += f'{bound[l]*180/pi: >13.3f}°     {delim[l]}'
                    else:
                        line += f'{bound[l]: >13.3f}      {delim[l]}'
                print(line)
            print(3*' '+61*'-')
        # compute velocity quantiles
        def velocity_quantiles(self,quantiles=array([0.16,0.84]),verbose=True):
            uscl = 101
            uarr = array(list(map(lambda i: float(i)/uscl,range(-uscl,uscl+1,1))))
            psi  = uarr*self.params['TerminalVelocity'] # psi = u * vterm/c
            wave = self.w0[0] * sqrt( (1+psi) / (1-psi) ) # longitudinal D-shift
            vels = psi*2.99792458e5 # convert from c units to km / s
            if self.settings['Profile'] != 'Absorption' :
                cdf  = cumulative_trapezoid(self.get_profile(wave),x=wave)
            elif self.settings['Profile'] == 'Absorption' :
                cdf  = cumulative_trapezoid(1-self.get_profile(wave),x=wave)
            else:
                print('Velocity Quantiles Not Supported for P Cygni.')
                print('Try again using Nebular or Absorption profiles')
                print('with the same parameters.')
                return nan
            cdf /= cdf[-1]
            vel_quant = interp(quantiles,cdf,vels[1:])
            if verbose:
                print(3*' '+34*'-'+'\n  |'+8*' '+'VELOCITY QUANTILES'+8*' '+'|\n'+3*' '+34*'-')
                for qi,vq in zip(quantiles,vel_quant):
                    print('  |   '+f'{qi: >8.6f}  :  {vq: >8.3f} km s^-1  |')
                print(3*' '+34*'-')
            return vel_quant
        # convenience functions for different common velocity quantiles
        def get_W80(self):
            vq = self.velocity_quantiles(quantiles=[0.1,0.9],verbose=False)
            return vq[1]-vq[0]
        def get_W50(self):
            vq = self.velocity_quantiles(quantiles=[0.25,0.75],verbose=False)
            return vq[1]-vq[0]
        def get_FWHF(self):
            return self.get_W50()
        def get_1sigma(self):
            vq = self.velocity_quantiles(quantiles=[0.1587,0.8413],verbose=False)
            return vq[1]-vq[0]
        # initiate parameters
        def init_params(self,disp=2**-6):
            nwalk = 2*self.npar+1
            init = self.get_params()*(1 + disp*randn(nwalk,self.npar))
            for i in range(nwalk):
                for j in range(self.npar):
                    while init[i,j] < self.get_bounds()[0][j] or \
                                            init[i,j] > self.get_bounds()[1][j]:
                        init[i,j] = self.get_params()[j]*(1 + disp*randn())
                if self.settings['Geometry'] == 'HollowCones' :
                    j = where(self.get_param_names()=='OpeningAngle')[0]
                    while init[i,j+1] >= init[i,j] - 0.0872664626 :
                        init[i,j+1] = init[i,j] - 0.0872664626+disp*randn()
            return init,nwalk
        # log likelihood for chi squared
        def log_likelihood(self,theta,x,y,yerr):
            ybar = self.Profile(x,*theta)
            ln_prb = -0.5*((y-ybar)/(yerr))**2-2*log(yerr)
            return nansum(ln_prb)
        # boundary condition check
        @staticmethod
        def __fun_bound__(t,tl,tu):
            return (tl<t)&(t<tu)&(isfinite(t))
        # log prior -- uniform bounded
        def log_prior(self,theta):
            tlower,tupper = self.get_bounds()
            tbound = list(map(self.__fun_bound__,theta,tlower,tupper))
            if all(tbound):
                return 0
            else:
                return -inf
        # log probability -- prior plus likelihood
        def log_probability(self,theta,x,y,var):
            lp = self.log_prior(theta)
            if not isfinite(lp):
                return -inf
            else:
                ll = self.log_likelihood(theta,x,y,var)
                if not isfinite(ll):
                    return -inf
                else:
                    return lp + ll

    return Profile
# various "parent" line profile classes
@profile_constructor
class Nebular():
    '''
    Name:
        Nebular

    Purpose:
        Compute a model nebular emission line profile for outflowing gas for the
        given atomic data and assumed velocity field, geometry, and density profile.

    Arguments:
        :w0 (*float*): rest cental wavelength of the line

    Keyword arguments:
            :VelocityField (*str*): radial velocity field indicating the assumed
                    treatment of the underlying physics and acceleration.
                    Options include
                         \'BetaCAK\' (1-1/x)^b from CAK theory
                         \'AccPlaw\' sqrt(1-x^(1-b)) from Steidel et al. 2010
                         \'VelPlaw\' 0.5(x-1)^b from various sources
            :Geometry (*str*): azimuthal geometry capturing the (an)isotropy
                    Options include
                        \'Sphere\' or \'Spherical\' isotropic case
                        \'FilledCones\' bidirectional cones
                        \'HollowCones\' or \'OpenCones\' cones with a cavity
                        \'HollowConesFixedCavity\' cones with fixed cavity size
            :DensityProfile (*str*): radial profile to use for the gas density.
                    Options include integration of many pulses or episodes
                         \'PowerLaw\'
                         \'Exponential\'
                         \'PowerLaw2\'
                     and individual pulse-like episodes or outbursts
                         \'LogNormal\'
                         \'Normal\'
                         \'Shell\'
                         \'FRED\'
                    or a combination of the two
                         \'Pulses\'
                         \'DampedPulses\'
            :Disk     (*bool*): boolean indicating whether a disk is present and
                    obstructs the cone posterior to the observer -- requires cone
                    geometry in order to take effect
            :Aperture (*bool*): boolean indicating whether the observed profile is
                    truncated due to the effects of a spherical aperture

    Returns:
        :Profile (*object*): An instance of the Profile object class inhereting
                the Nebular class with properties specific to an emission line

    Attributes:
        :Profile (*function*): function returning a line profile based on the chosen
                velocity field, geometry, and density profile; takes wavelength and
                paramaters (see the `params` attribute)
        :params (*dict*): dictionary of parameters, initially assigned or later
                updated by the user
        :npar (*int*): number of assigned parameters
        :bounds (*dict*): dictionary of parameter bounds, initially assigned or
                later updated by the user

    Bound methods:
        :get_params: returns the current parameters
        :update_params: updates parameter(s) given the name(s) and value(s)
        :get_bounds: returns the current boundaries set for the parameters
        :update_bounds: updates bound(s) given the name(s) and value(s)
        :print_params: prints current parameter values to the command line

    '''
    def __init__(self,w0):
        kwargs = {k:self.settings[k] for k in \
            ['VelocityField','DensityProfile','Geometry','Disk','FromRest','Aperture']}
        self.OutLinesModel = build_profile_model('Nebular',**kwargs)
        pass
    # define profile with and without a static ISM component
    # order of parameters:
    # static flux(es), Doppler width, outflow flux(es), terminal velocity
    # velocity field index, geometry terms, density profile terms
    def __WithStatic__(self,w,sigv,vinf,beta,*pars):
        Fs = pars[:self.nLines]
        Fo = pars[self.nLines:2*self.nLines]
        po = pars[2*self.nLines:]
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*po), self.w0 )))
        static  = array(list(map(lambda w0: gauss(w,w0,sigv), self.w0 )))
        return Fs@static + Fo@OutLine
    def __WithoutStatic__(self,w,vinf,beta,*pars):
        Fo = pars[:self.nLines]
        po = pars[self.nLines:]
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*po), self.w0 )))
        return Fo@OutLine
    def __OnlyStatic__(self,w,sigv,*Fs):
        static  = array(list(map(lambda w0: gauss(w,w0,sigv), self.w0 )))
        return Fs@static
@profile_constructor
class Absorption():
    '''
    Name:
        Absorption

    Purpose:
        Compute a model spectral absorption line profile for outflowing gas for the
        given atomic data and assumed velocity field, geometry, and density profile.

    Arguments:
        :w0 (*float*): rest cental wavelength of the line
        :fosc (*float*): oscilator strength of the line

    Keyword arguments:
            :VelocityField (*str*): radial velocity field indicating the assumed
                    treatment of the underlying physics and acceleration.
                    Options include
                         \'BetaCAK\' (1-1/x)^b from CAK theory
                         \'AccPlaw\' sqrt(1-x^(1-b)) from Steidel et al. 2010
                         \'VelPlaw\' 0.5(x-1)^b from various sources
            :Geometry (*str*): azimuthal geometry capturing the (an)isotropy
                    Options include
                        \'Sphere\' or \'Spherical\' isotropic case
                        \'FilledCones\' bidirectional cones
                        \'HollowCones\' or \'OpenCones\' cones with a cavity
                        \'HollowConesFixedCavity\' cones with fixed cavity size
            :DensityProfile (*str*): radial profile to use for the gas density.
                    Options include integration of many pulses or episodes
                         \'PowerLaw\'
                         \'Exponential\'
                         \'PowerLaw2\'
                     and individual pulse-like episodes or outbursts
                         \'LogNormal\'
                         \'Normal\'
                         \'Shell\'
                         \'FRED\'
                    or a combination of the two
                         \'Pulses\'
                         \'DampedPulses\'
            :Disk     (*bool*): boolean indicating whether a disk is present and
                    obstructs the cone posterior to the observer -- requires cone
                    geometry in order to take effect
            :Aperture (*bool*): boolean indicating whether the observed profile is
                    truncated due to the effects of a spherical aperture

    Returns:
        :Profile (*object*): An instance of the Profile object class inhereting
                the Absorption class with properties specific to an absorption line

    Attributes:
        :Profile (*function*): function returning a line profile based on the chosen
                velocity field, geometry, and density profile; takes wavelength and
                paramaters (see the `params` attribute)
        :params (*dict*): dictionary of parameters, initially assigned or later
                updated by the user
        :npar (*int*): number of assigned parameters
        :bounds (*dict*): dictionary of parameter bounds, initially assigned or
                later updated by the user

    Bound methods:
        :get_params: returns the current parameters
        :update_params: updates parameter(s) given the name(s) and value(s)
        :get_bounds: returns the current boundaries set for the parameters
        :update_bounds: updates bound(s) given the name(s) and value(s)
        :print_params: prints current parameter values to the command line

    '''
    def __init__(self,w0,fosc):
        kwargs = {k:self.settings[k] for k in \
            ['VelocityField','DensityProfile','Geometry','Disk','FromRest','Aperture']}
        self.OutLinesModel = build_profile_model('Absorption',**kwargs)
        pass
    # define profile with and without a static ISM component
    def __WithStatic__(self,w,sigv,vinf,beta,*pars):
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*pars[2*self.nLines:]),self.w0)))
        static  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        logNo = array(pars[self.nLines:2*self.nLines])
        taus = -self.w0*self.fosc*10**(logNs-14.8247)/(sigv*2.99792458e5)
        tauo = -self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        return exp( taus@static + tauo@OutLine )
    def __WithoutStatic__(self,w,vinf,beta,*pars):
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*pars[self.nLines:]),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNo = array(pars[:self.nLines])
        tauo = -self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        return exp( tauo@OutLine )
    def __OnlyStatic__(self,w,sigv,*pars):
        static  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        taus = -self.w0*self.fosc*10**(logNs-14.8247)/(sigv*2.99792458e5)
        return exp( taus@static )
@profile_constructor
class Resonant():
    '''
    Name:
        Resonant

    Purpose:
        Compute a model spectral resonant line profile for
        outflowing gas for the given atomic data and assumed velocity field,
        geometry, and density profile.

    Arguments:
        :w0 (*float*): rest cental wavelength of the line
        :fosc (*float*): oscilator strength of the line
        :pline (*float*): relative spontaneous transition probability of the
                    line through the resonant channel:
                        p_r = A_r / (A_f + A_r)

    Keyword arguments:
            :VelocityField (*str*): radial velocity field indicating the assumed
                    treatment of the underlying physics and acceleration.
                    Options include
                         \'BetaCAK\' (1-1/x)^b from CAK theory
                         \'AccPlaw\' sqrt(1-x^(1-b)) from Steidel et al. 2010
                         \'VelPlaw\' 0.5(x-1)^b from various sources
            :Geometry (*str*): azimuthal geometry capturing the (an)isotropy
                    Options include
                        \'Sphere\' or \'Spherical\' isotropic case
                        \'FilledCones\' bidirectional cones
                        \'HollowCones\' or \'OpenCones\' cones with a cavity
                        \'HollowConesFixedCavity\' cones with fixed cavity size
            :DensityProfile (*str*): radial profile to use for the gas density.
                    Options include integration of many pulses or episodes
                         \'PowerLaw\'
                         \'Exponential\'
                         \'PowerLaw2\'
                     and individual pulse-like episodes or outbursts
                         \'LogNormal\'
                         \'Normal\'
                         \'Shell\'
                         \'FRED\'
                    or a combination of the two
                         \'Pulses\'
                         \'DampedPulses\'
            :Disk     (*bool*): boolean indicating whether a disk is present and
                    obstructs the cone posterior to the observer -- requires cone
                    geometry in order to take effect
            :Aperture (*bool*): boolean indicating whether the observed profile is
                    truncated due to the effects of a spherical aperture

    Returns:
        :Profile (*object*): An instance of the Profile object class inhereting
                the Absorption class with properties specific to an absorption line

    Attributes:
        :Profile (*function*): function returning a line profile based on the chosen
                velocity field, geometry, and density profile; takes wavelength and
                paramaters (see the `params` attribute)
        :params (*dict*): dictionary of parameters, initially assigned or later
                updated by the user
        :npar (*int*): number of assigned parameters
        :bounds (*dict*): dictionary of parameter bounds, initially assigned or
                later updated by the user

    Bound methods:
        :get_params: returns the current parameters
        :update_params: updates parameter(s) given the name(s) and value(s)
        :get_bounds: returns the current boundaries set for the parameters
        :update_bounds: updates bound(s) given the name(s) and value(s)
        :print_params: prints current parameter values to the command line

    '''
    def __init__(self,w0,fosc,pline):
        kwargs = {k:self.settings[k] for k in \
            ['VelocityField','DensityProfile','Geometry','Disk','FromRest','Aperture']}
        self.OutLinesModel = build_profile_model('Resonant',**kwargs)
        pass
    # define profile with and without a static ISM component
    def __WithStatic__(self,w,sigv,vinf,beta,*pars):
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*pars[2*self.nLines:]),self.w0)))
        staticv  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        #staticg  = array(list(map(lambda w0: gauss(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        logNo = array(pars[self.nLines:2*self.nLines])
        # static
        tausr = self.w0*self.fosc*10**(logNs-14.8247)/(vinf*2.99792458e5)
        fescs = array(list(map(self.__EscapeFraction__,tausr,self.pline,staticv)))
        srcfs = array(list(map(self.__SourceFunction__,tausr,staticv)))
        # OutLine
        tauor = self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        fesco = array(list(map(self.__EscapeFraction__,tauor,self.pline,OutLine)))
        srcfo = array(list(map(self.__SourceFunction__,tauor,OutLine)))
        return (fesco*srcfo).sum(axis=0) + (fescs*srcfs).sum(axis=0)
    def __WithoutStatic__(self,w,vinf,beta,*pars):
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*pars[self.nLines:]),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light c = 2.99792458e5 km/s
        logNo = array(pars[:self.nLines])
        tauor = self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        fesco = array(list(map(self.__EscapeFraction__,tauor,self.pline,OutLine)))
        srcfo = array(list(map(self.__SourceFunction__,tauor,OutLine)))
        return (fesco*srcfo).sum(axis=0)
    def __OnlyStatic__(self,w,sigv,*pars):
        staticv  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        #staticg  = array(list(map(lambda w0: gauss(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        tausr = self.w0*self.fosc*10**(logNs-14.8247)/(vinf*2.99792458e5)
        fescs = array(list(map(self.__EscapeFraction__,tausr,self.pline,staticv)))
        srcfs = array(list(map(self.__SourceFunction__,tausr,staticv)))
        return (fescs*srcfs).sum(axis=0)
    @staticmethod
    def __EscapeFraction__(tau,pline,profile):
        pesc = (1-exp(-tau*profile))/(tau*profile)
        #pesc[isnan(pesc)] = 1
        fesc = pline*pesc/(1-pline*(1-pesc))
        fesc[profile<2**-24] = pline
        return fesc
    @staticmethod
    def __SourceFunction__(tau,profile):
        return 1-exp(-tau*profile)
@profile_constructor
class Fluorescent():
    '''
    Name:
        Fluorescent

    Purpose:
        Compute a model spectral fluorescent line profile for
        outflowing gas for the given atomic data and assumed velocity field,
        geometry, and density profile.

    Arguments:
        :w0 (*float*): rest cental wavelength of the line
        :fosc (*float*): oscilator strength of the line
        :pline (*float*): relative spontaneous transition probability of the
                    line through the fluorescent channel:
                        p_f = A_f / (A_f + A_r)

    Keyword arguments:
            :VelocityField (*str*): radial velocity field indicating the assumed
                    treatment of the underlying physics and acceleration.
                    Options include
                         \'BetaCAK\' (1-1/x)^b from CAK theory
                         \'AccPlaw\' sqrt(1-x^(1-b)) from Steidel et al. 2010
                         \'VelPlaw\' 0.5(x-1)^b from various sources
            :Geometry (*str*): azimuthal geometry capturing the (an)isotropy
                    Options include
                        \'Sphere\' or \'Spherical\' isotropic case
                        \'FilledCones\' bidirectional cones
                        \'HollowCones\' or \'OpenCones\' cones with a cavity
                        \'HollowConesFixedCavity\' cones with fixed cavity size
            :DensityProfile (*str*): radial profile to use for the gas density.
                    Options include integration of many pulses or episodes
                         \'PowerLaw\'
                         \'Exponential\'
                         \'PowerLaw2\'
                     and individual pulse-like episodes or outbursts
                         \'LogNormal\'
                         \'Normal\'
                         \'Shell\'
                         \'FRED\'
                    or a combination of the two
                         \'Pulses\'
                         \'DampedPulses\'
            :Disk     (*bool*): boolean indicating whether a disk is present and
                    obstructs the cone posterior to the observer -- requires cone
                    geometry in order to take effect
            :Aperture (*bool*): boolean indicating whether the observed profile is
                    truncated due to the effects of a spherical aperture

    Returns:
        :Profile (*object*): An instance of the Profile object class inhereting
                the Absorption class with properties specific to an absorption line

    Attributes:
        :Profile (*function*): function returning a line profile based on the chosen
                velocity field, geometry, and density profile; takes wavelength and
                paramaters (see the `params` attribute)
        :params (*dict*): dictionary of parameters, initially assigned or later
                updated by the user
        :npar (*int*): number of assigned parameters
        :bounds (*dict*): dictionary of parameter bounds, initially assigned or
                later updated by the user

    Bound methods:
        :get_params: returns the current parameters
        :update_params: updates parameter(s) given the name(s) and value(s)
        :get_bounds: returns the current boundaries set for the parameters
        :update_bounds: updates bound(s) given the name(s) and value(s)
        :print_params: prints current parameter values to the command line

    '''
    def __init__(self,w0,fosc_flu,fosc_res,pline):
        kwargs = {k:self.settings[k] for k in \
            ['VelocityField','DensityProfile','Geometry','Disk','FromRest','Aperture']}
        self.OutLinesModel = build_profile_model('Fluorescent',**kwargs)
        pass
    # define profile with and without a static ISM component
    def __WithStatic__(self,w,sigv,vinf,beta,*pars):
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*pars[2*self.nLines:]),self.w0)))
        staticv  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        #staticg  = array(list(map(lambda w0: gauss(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        logNo = array(pars[self.nLines:2*self.nLines])
        # static
        tausr = self.w0*self.fres*10**(logNs-14.8247)/(vinf*2.99792458e5)
        tausf = self.w0*self.fosc*10**(logNs-14.8247)/(vinf*2.99792458e5)
        fescs = array(list(map(self.__EscapeFraction__,tausr,self.pline,staticv)))
        srcfs = array(list(map(self.__SourceFunction__,tausf,staticv)))
        # OutLine
        tauor = self.w0*self.fres*10**(logNo-14.8247)/(vinf*2.99792458e5)
        tauof = self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        fesco = array(list(map(self.__EscapeFraction__,tauor,self.pline,OutLine)))
        srcfo = array(list(map(self.__SourceFunction__,tauof,OutLine)))
        return (fesco*srcfo).sum(axis=0) + (fescs*srcfs).sum(axis=0)
    def __WithoutStatic__(self,w,vinf,beta,*pars):
        OutLine = array(list(map(lambda w0: \
          self.OutLinesModel(w,w0,vinf,beta,*pars[self.nLines:]),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light c = 2.99792458e5 km/s
        logNo = array(pars[:self.nLines])
        tauor = self.w0*self.fres*10**(logNo-14.8247)/(vinf*2.99792458e5)
        tauof = self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        fesco = array(list(map(self.__EscapeFraction__,tauor,self.pline,OutLine)))
        srcfo = array(list(map(self.__SourceFunction__,tauof,OutLine)))
        return (fesco*srcfo).sum(axis=0)
    def __OnlyStatic__(self,w,sigv,*pars):
        staticv  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        #staticg  = array(list(map(lambda w0: gauss(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        tausr = self.w0*self.fres*10**(logNs-14.8247)/(vinf*2.99792458e5)
        tausf = self.w0*self.fosc*10**(logNs-14.8247)/(vinf*2.99792458e5)
        fescs = array(list(map(self.__EscapeFraction__,tausr,self.pline,staticv)))
        srcfs = array(list(map(self.__SourceFunction__,tausf,staticv)))
        return (fescs*srcfs).sum(axis=0)
    @staticmethod
    def __EscapeFraction__(tau,pline,profile):
        pesc = (1-exp(-tau*profile))/(tau*profile)
        fesc = pline/(1-(1-pline)*(1-pesc))
        fesc[profile<2**-24] = pline
        return fesc
    @staticmethod
    def __SourceFunction__(tau,profile):
        return 1-exp(-tau*profile)
@profile_constructor
class PCygni():
    '''
    Name:
        PCygni

    Purpose:
        Compute a model spectral resonant P Cygni line profile for
        outflowing gas for the given atomic data and assumed velocity field,
        geometry, and density profile.

    Arguments:
        :w0 (*float*): rest cental wavelength of the line
        :fosc (*float*): oscilator strength of the line

    Keyword arguments:
            :VelocityField (*str*): radial velocity field indicating the assumed
                    treatment of the underlying physics and acceleration.
                    Options include
                         \'BetaCAK\' (1-1/x)^b from CAK theory
                         \'AccPlaw\' sqrt(1-x^(1-b)) from Steidel et al. 2010
                         \'VelPlaw\' 0.5(x-1)^b from various sources
            :Geometry (*str*): azimuthal geometry capturing the (an)isotropy
                    Options include
                        \'Sphere\' or \'Spherical\' isotropic case
                        \'FilledCones\' bidirectional cones
                        \'HollowCones\' or \'OpenCones\' cones with a cavity
                        \'HollowConesFixedCavity\' cones with fixed cavity size
            :DensityProfile (*str*): radial profile to use for the gas density.
                    Options include integration of many pulses or episodes
                         \'PowerLaw\'
                         \'Exponential\'
                         \'PowerLaw2\'
                     and individual pulse-like episodes or outbursts
                         \'LogNormal\'
                         \'Normal\'
                         \'Shell\'
                         \'FRED\'
                    or a combination of the two
                         \'Pulses\'
                         \'DampedPulses\'
            :Disk     (*bool*): boolean indicating whether a disk is present and
                    obstructs the cone posterior to the observer -- requires cone
                    geometry in order to take effect
            :Aperture (*bool*): boolean indicating whether the observed profile is
                    truncated due to the effects of a spherical aperture

    Returns:
        :Profile (*object*): An instance of the Profile object class inhereting
                the Absorption class with properties specific to an absorption line

    Attributes:
        :Profile (*function*): function returning a line profile based on the chosen
                velocity field, geometry, and density profile; takes wavelength and
                paramaters (see the `params` attribute)
        :params (*dict*): dictionary of parameters, initially assigned or later
                updated by the user
        :npar (*int*): number of assigned parameters
        :bounds (*dict*): dictionary of parameter bounds, initially assigned or
                later updated by the user

    Bound methods:
        :get_params: returns the current parameters
        :update_params: updates parameter(s) given the name(s) and value(s)
        :get_bounds: returns the current boundaries set for the parameters
        :update_bounds: updates bound(s) given the name(s) and value(s)
        :print_params: prints current parameter values to the command line

    '''
    def __init__(self,w0,fosc,pline):
        kwargs = {k:self.settings[k] for k in \
            ['VelocityField','DensityProfile','Geometry','Disk','FromRest','Aperture']}
        self.OutLinesModelA = __absorption__.build_profile_model('Absorption',**kwargs)
        self.OutLinesModelR = __resonfluor__.build_profile_model('Resonant',**kwargs)
        pass
    # define profile with and without a static ISM component
    def __WithStatic__(self,w,sigv,vinf,beta,*pars):
        OutLineA = array(list(map(lambda w0: \
          self.OutLinesModelA(w,w0,vinf,beta,*pars[2*self.nLines:]),self.w0)))
        OutLineR = array(list(map(lambda w0: \
          self.OutLinesModelR(w,w0,vinf,beta,*pars[2*self.nLines:]),self.w0)))
        staticv  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        #staticg  = array(list(map(lambda w0: gauss(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        logNo = array(pars[self.nLines:2*self.nLines])
        # static
        tausr = self.w0*self.fosc*10**(logNs-14.8247)/(vinf*2.99792458e5)
        fescs = array(list(map(self.__EscapeFraction__,tausr,self.pline,staticv)))
        srcfs = array(list(map(self.__SourceFunction__,tausr,staticv)))
        # OutLine
        tauor = self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        fesco = array(list(map(self.__EscapeFraction__,tauor,self.pline,OutLineR)))
        srcfo = array(list(map(self.__SourceFunction__,tauor,OutLineR)))
        return (fesco*srcfo).sum(axis=0) + (fescs*srcfs).sum(axis=0) + \
                exp( - tausr@staticv - tauor@OutLineA )
    def __WithoutStatic__(self,w,vinf,beta,*pars):
        OutLineA = array(list(map(lambda w0: \
          self.OutLinesModelA(w,w0,vinf,beta,*pars[2*self.nLines:]),self.w0)))
        OutLineR = array(list(map(lambda w0: \
          self.OutLinesModelR(w,w0,vinf,beta,*pars[2*self.nLines:]),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light c = 2.99792458e5 km/s
        logNo = array(pars[:self.nLines])
        tauor = self.w0*self.fosc*10**(logNo-14.8247)/(vinf*2.99792458e5)
        fesco = array(list(map(self.__EscapeFraction__,tauor,self.pline,OutLineR)))
        srcfo = array(list(map(self.__SourceFunction__,tauor,OutLineR)))
        return (fesco*srcfo).sum(axis=0) + exp( - tauor@OutLineA )
    def __OnlyStatic__(self,w,sigv,*pars):
        staticv  = array(list(map(lambda w0: voigt(w,w0,sigv),self.w0)))
        #staticg  = array(list(map(lambda w0: gauss(w,w0,sigv),self.w0)))
        # classical absorption cross-section
        # log(sigma) = -14.8247 in cm^2, b in km/s, w in Ang
        # speed of light
        # c = 2.99792458e5 km/s
        logNs = array(pars[:self.nLines])
        tausr = self.w0*self.fosc*10**(logNs-14.8247)/(vinf*2.99792458e5)
        fescs = array(list(map(self.__EscapeFraction__,tausr,self.pline,staticv)))
        srcfs = array(list(map(self.__SourceFunction__,tausr,staticv)))
        return (fescs*srcfs).sum(axis=0) + exp( - tausr@staticv)
    @staticmethod
    def __EscapeFraction__(tau,pline,profile):
        pesc = (1-exp(-tau*profile))/(tau*profile)
        fesc = pline*pesc/(1-pline*(1-pesc))
        fesc[profile<2**-24] = pline
        return fesc
    @staticmethod
    def __SourceFunction__(tau,profile):
        return 1-exp(-tau*profile)

#
'''
Name:
    build_profile_model

Purpose:
    Construct a unique function based on user-specified model options. This
    function will then compute and return the normalized line profiles.

Keyword Arguments:
            :VelocityField (*str*): radial velocity field indicating the assumed
                    treatment of the underlying physics and acceleration.
                    Options include
                         \'BetaCAK\' (1-1/x)^b from CAK theory
                         \'AccPlaw\' sqrt(1-x^(1-b)) from Steidel et al. 2010
                         \'VelPlaw\' 0.5(x-1)^b from various sources
            :Geometry (*str*): azimuthal geometry capturing the (an)isotropy
                    Options include
                        \'Sphere\' or \'Spherical\' isotropic case
                        \'FilledCones\' bidirectional cones
                        \'HollowCones\' or \'OpenCones\' cones with a cavity
                        \'HollowConesFixedCavity\' cones with fixed cavity size
            :DensityProfile (*str*): radial profile to use for the gas density.
                    Options include integration of many pulses or episodes
                         \'PowerLaw\'
                         \'Exponential\'
                         \'PowerLaw2\'
                     and individual pulse-like episodes or outbursts
                         \'LogNormal\'
                         \'Normal\'
                         \'Shell\'
                         \'FRED\'
                    or a combination of the two
                         \'Pulses\'
                         \'DampedPulses\'
            :Disk     (*bool*): boolean indicating whether a disk is present and
                    obstructs the cone posterior to the observer -- requires cone
                    geometry in order to take effect
            :Aperture (*bool*): boolean indicating whether the observed profile is
                    truncated due to the effects of a spherical aperture

Returns:
    :profile (*function*): a function which calculates the line profile for a
                        galactic outflow for the user-provided model settings

'''
def build_profile_model(LineType,VelocityField='BetaCAK',DensityProfile='PowerLaw',Geometry='Sphere',Pulse='Normal',Aperture=False,Disk=False,FromRest=True):
    # possible profile model choices
    kwargs = dict(VF=VelocityField,DP=DensityProfile,Pulse=Pulse)
    # different line profile types
    calc_phi = {'Nebular':      __nebular__.calc_phi,\
                'Resonant':     __resonfluor__.calc_phi,\
                'Fluorescent':  __resonfluor__.calc_phi,\
                'Absorption':   __absorption__.calc_phi,\
                }
    ###
    ### All possible profile geometries with launching from rest
    ###
    if FromRest:
        # hemisphere geometry
        if 'spher' in Geometry.lower() and 'hemi' in Geometry.lower():
            if Aperture :
                def profile(w,w0,vinf,beta,incl,vapr,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,pi/2,0,inf,1e-7,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            else:
                def profile(w,w0,vinf,beta,incl,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,pi/2,0,inf,1e-7,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
        # sphere geometry
        elif 'spher' in Geometry.lower() and 'hemi' not in Geometry.lower():
            if Aperture :
                def profile(w,w0,vinf,beta,vapr,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,0,pi/2,0,0,1e-7,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            else:
                def profile(w,w0,vinf,beta,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,0,pi/2,0,0,1e-7,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
        # filled cone geometry
        elif 'filled' in Geometry.lower():
            if Aperture and Disk :
                def profile(w,w0,vinf,beta,incl,tO,xdisk,vapr,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,0,xdisk,1e-7,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            elif Disk and not Aperture :
                def profile(w,w0,vinf,beta,incl,tO,xdisk,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,0,xdisk,1e-7,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            elif Aperture and not Disk :
                def profile(w,w0,vinf,beta,incl,tO,vapr,*par):
                    phi = calc_phi[LineType](w,w0,vinf,incl,tO,0,0,1e-7,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            else:
                def profile(w,w0,vinf,beta,incl,tO,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,0,0,1e-7,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
        # open cone geometry
        elif 'hollow' in Geometry.lower() or 'cavity' in Geometry.lower() or 'open' in Geometry.lower():
            if 'fix' in Geometry.lower():
                if Aperture and Disk :
                    def profile(w,w0,vinf,beta,incl,tO,xdisk,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,xdisk,1e-7,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Disk and not Aperture :
                    def profile(w,w0,vinf,beta,incl,tO,xdisk,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,xdisk,1e-7,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Aperture and not Disk :
                    def profile(w,w0,vinf,beta,incl,tO,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,0,1e-7,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                else:
                    def profile(w,w0,vinf,beta,incl,tO,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,0,1e-7,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
            else:
                if Aperture and Disk :
                    def profile(w,w0,vinf,beta,incl,tO,tC,xdisk,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,xdisk,1e-7,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Disk and not Aperture :
                    def profile(w,w0,vinf,beta,incl,tO,tC,xdisk,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,xdisk,1e-7,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Aperture and not Disk :
                    def profile(w,w0,vinf,beta,incl,tO,tC,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,0,1e-7,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                else:
                    def profile(w,w0,vinf,beta,incl,tO,tC,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,0,1e-7,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
        else:
            print('Geometry not recognized. Options are \'Sphere\',\n'+
                '\'Hemisphere\', \'FilledCones\', \'HollowCones\', or \'HollowConesFixedCavity\'.')
    ###
    ### All possible profile geometries with launching from rest
    ###
    else:
        if 'spher' in Geometry.lower() and 'hemi' in Geometry.lower():
            if Aperture :
                def profile(w,w0,vinf,beta,incl,vini,vapr,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,pi/2,0,inf,vini,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            else:
                def profile(w,w0,vinf,beta,incl,vini,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,pi/2,0,inf,vini,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
        elif 'spher' in Geometry.lower() and 'hemi' not in Geometry.lower():
            if Aperture :
                def profile(w,w0,vinf,beta,vini,vapr,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,0,pi/2,0,0,vini,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            else:
                def profile(w,w0,vinf,beta,vini,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,0,pi/2,0,0,vini,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
        elif 'filled' in Geometry.lower():
            if Aperture and Disk :
                def profile(w,w0,vinf,beta,incl,tO,xdisk,vini,vapr,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,0,xdisk,vini,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            elif Disk and not Aperture :
                def profile(w,w0,vinf,beta,incl,tO,xdisk,vini,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,0,xdisk,vini,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            elif Aperture and not Disk :
                def profile(w,w0,vinf,beta,incl,tO,vini,vapr,*par):
                    phi = calc_phi[LineType](w,w0,vinf,incl,tO,0,0,vini,vapr,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
            else:
                def profile(w,w0,vinf,beta,incl,tO,vini,*par):
                    phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,0,0,vini,1,*par,**kwargs)
                    return phi/trapezoid(phi,x=w)
        elif 'hollow' in Geometry.lower() or 'cavity' in Geometry.lower() or 'open' in Geometry.lower():
            if 'fix' in Geometry.lower():
                if Aperture and Disk :
                    def profile(w,w0,vinf,beta,incl,tO,xdisk,vini,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,xdisk,vini,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Disk and not Aperture :
                    def profile(w,w0,vinf,beta,incl,tO,xdisk,vini,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,xdisk,vini,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Aperture and not Disk :
                    def profile(w,w0,vinf,beta,incl,tO,vini,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,0,vini,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                else:
                    def profile(w,w0,vinf,beta,incl,tO,vini,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tO-0.174533,0,vini,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
            else:
                if Aperture and Disk :
                    def profile(w,w0,vinf,beta,incl,tO,tC,xdisk,vini,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,xdisk,vini,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Disk and not Aperture :
                    def profile(w,w0,vinf,beta,incl,tO,tC,xdisk,vini,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,xdisk,vini,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                elif Aperture and not Disk :
                    def profile(w,w0,vinf,beta,incl,tO,tC,vini,vapr,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,0,vini,vapr,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
                else:
                    def profile(w,w0,vinf,beta,incl,tO,tC,vini,*par):
                        phi = calc_phi[LineType](w/w0,vinf,beta,incl,tO,tC,0,vini,1,*par,**kwargs)
                        return phi/trapezoid(phi,x=w)
        else:
            print('Geometry not recognized. Options are \'Sphere\',\n'+
                '\'Hemisphere\', \'FilledCones\', \'HollowCones\', or \'HollowConesFixedCavity\'.')
    return profile
'''
Name:
    profile

Purpose:
    Calculate the spectral line profile
    under the Sobolev approximation for the user-specified geometry,
    density profile, and velocity field.

Arguments:
    :wave   (*np.ndarray*) : array of observed wavelengths
    :wave0  (*float*) : rest-frame central wavelength
    :vinf   (*float*) : terminal velocity in c units
    :beta   (*float*) : power-law index of velocity field
    :incl   (*float*) : inclination of the cone
                        (only if filled or hollow conical Geometry)
    :thetaO (*float*) : opening angle of the filled cone
                        (only if filled or hollow conical Geometry)
    :thetaC (*float*) : opening angle of cavity in cone
                        (only if hollow cone Geometry)
    :xdisk  (*float*) : scaled disk radius (only provide if Disk)
    :vapr   (*float*) : scaled aperture limit on velocity
                        (only if correcting for a circular aperture)
    :p1,p2,...: (*floats*) : density field parameters, must match selected
                        'DP' -- see documentation

Returns:
    :profile (*function*) : function which computes the normalized line profile
                        in units of wavelength^(-1) for the given options
'''



from numpy import array,append,ones,zeros,absolute,where,max,min,sum,nansum,\
                    log,log10,exp,sqrt,cos,sin,arccos,arcsin,pi,inf,nan,isnan,\
                    isfinite,sign,arange,linspace,logspace,argmin,nanmin,\
                    argmax,nanmax,interp,diff,quantile,nanquantile,square
from numpy.random import seed,default_rng
global uniform_sampler
uniform_sampler = default_rng().random
from scipy.optimize import brentq,newton,fminbound
from scipy.integrate import trapezoid,simpson,fixed_quad,cumulative_trapezoid
from scipy.special import wofz
from functools import partial
###
### Density Profiles
###
##
## WINDS
##
# power law
def dens_plaw(u,beta,vini,VF,alpha):
    return x[VF](u,beta,vini)**-alpha
# exponential
def dens_exp(u,beta,vini,VF,r):
    return exp(r-r*x[VF](u,beta,vini))
# double power law
def dens_dplaw(u,beta,vini,VF,a1,a2,x1):
    xuv = x[VF](u,beta,vini)
    if hasattr(u,'__len__'):
        den = zeros(len(u))
        den[xuv < x1]  = xuv[xuv < x1]**-a1
        den[xuv >= x1] = x1**(a2-a1) * xuv[xuv >= x1]**-a2
        return den
    else:
        if xuv < x1 :
            return xuv**-a1
        else:
            return x1**(a2-a1) * xuv**-a2
##
## BUBBLES
##
# derivative of logistic column density
def dens_dlogic(u,beta,vini,VF,x0,k):
    expon = exp(-k*(x[VF](u,beta,vini)-x0))
    return k*expon/(1+expon)**2
# uniform shell
def dens_shell(u,beta,vini,VF,xi,sigma):
    if hasattr(u,'__len__'):
        den = zeros(len(u))
        den[ absolute( x[VF](u,beta,vini) - xi ) < sigma ] = 1
        return den
    else:
        if x[VF](u,beta,vini) >  xi - sigma and x[VF](u,beta,vini) < xi + sigma :
            return 1
        else:
            return 0
# fast rise, exponential decay
def dens_fred(u,beta,vini,VF,r1,r2,x1):
    dx = x[VF](u,beta,vini) - (x1-sqrt(r1*r2))
    if hasattr(u,'__len__'):
        den = zeros(len(u))
        den[dx>0] = exp(2*sqrt(r1/r2)) * exp(-r1/dx[dx>0]-dx[dx>0]/r2)
        return den
    else:
        if dx > 0:
            return exp(2*sqrt(r1/r2)) * exp(-r1/dx-dx/r2)
        else:
            return 0
# log normal
def dens_lognorm(u,beta,vini,VF,x1,sigma):
    return exp( -0.5*( log10(x[VF](u,beta,vini)) - x1 )**2/sigma**2 )
# normal
def dens_norm(u,beta,vini,VF,xi,sigma):
    return exp( -0.5*( x[VF](u,beta,vini) - xi )**2/sigma**2 )
##
## PULSE/SHELL ENSEMBLES
##
# ensemble with linear spacing
def dens_pulseslin(u,beta,vini,VF,sigma,dx,x0,pulse='Normal'):
    dens_pulse = n[pulse]
    return sum([dens_pulse(u,beta,vini,VF,x,sigma) for x in arange(x0,x[VF](0.99,beta,vini)+dx,dx)],axis=0)
# ensemble with log spacing
def dens_pulseslog(u,beta,vini,VF,sigma,dx,x0,pulse='Normal'):
    dens_pulse = n[pulse]
    return sum([dens_pulse(u,beta,vini,VF,x,sigma) for x in 10**arange(log10(x0),log10(x[VF](0.99,beta,vini))+dx,dx)],axis=0)
# damped ensemble with linear spacing
def dens_pulsedamplin(u,beta,vini,VF,r,sigma,dx,x0,pulse='Normal'):
    return dens_pulseslin(u,beta,vini,VF,sigma,dx,x0,pulse=pulse) * dens_exp(u,beta,vini,VF,r)
# damped ensemble with log spacing
def dens_pulsedamplog(u,beta,vini,VF,r,sigma,dx,x0,pulse='Normal'):
    return dens_pulseslog(u,beta,vini,VF,sigma,dx,x0,pulse=pulse) * dens_exp(u,beta,vini,VF,r)
# packet ensemble with linear spacing
def dens_pulsegrouplin(u,beta,vini,VF,s1,x1,sigma,dx,x0,pulse='Normal'):
    return dens_pulseslin(u,beta,vini,VF,sigma,dx,x0,pulse=pulse) * dens_norm(u,beta,vini,VF,x1,s1)
# packet ensemble with log spacing
def dens_pulsegrouplog(u,beta,vini,VF,s1,x1,sigma,dx,x0,pulse='Normal'):
    return dens_pulseslog(u,beta,vini,VF,sigma,dx,x0,pulse=pulse) * dens_norm(u,beta,vini,VF,x1,s1)
###
### Velocity Fields
###
##
## expressed as x = f(w)
##
# Castor-Lamers 1979 / Pauldrach 1986 beta-law approximation to CAK 1975
def x_cak(u,beta,vini):
    return ( 1-((u-vini)/(1-vini))**(1/beta) )**-1
#  velocity power law
def x_vplaw(u,beta,vini,A=0.5):
    return (((u-vini)/(1-vini))/A)**(1/beta)+1
# Steidel 2010 acceleration power law
def x_aplaw(u,beta,vini):
    return ( 1  -  ((u-vini)/(1-vini))**2 )**(1/(1-beta))
# Murray 2005 optically thick radiation pressure
def x_M2005(u,beta,vini):
    return exp( ((u-vini)/(1-vini))**2 )
# my own exponential law
def x_expon(u,beta,vini):
    return 1-log(1-(u-vini)/(1-vini))/beta
##
## expressed as w = f(x)
##
# Castor-Lamers 1979 / Pauldrach 1986 beta-law approximation to CAK 1975
def w_cak(xv,beta,vini):
    return (1-vini)*(1-1/xv)**beta + vini
#  velocity power law
def w_vplaw(xv,beta,vini,A=0.5):
    return A*(1-vini)*(xv-1)**beta + vini
# Steidel 2010 acceleration power law
def w_aplaw(xv,beta,vini):
    return (1-vini)*(1-xv**(1-beta))**0.5 + vini
# Murray 2005 optically thick radiation pressure
def w_M2005(xv,beta,vini):
    return (1-vini)*sqrt(log(x)) + vini
# my own exponential law
def w_expon(xv,beta,vini):
    return (1-vini)*(1-exp(-beta*(xv-1))) + vini
##
## related differentials -- velocity gradients
##
# generalized callable
def dxdw(w,beta,vini,VF):
    return dxdv[VF](w,beta,vini)
# inverse of the radial velocity gradient
# via central finite difference method
def dxdw_cfd(w,beta,VF,h=2**-20):
    return (x[VF](w+h,beta)-x[VF](w,beta))/h
# explicit inverse of the radial velocity gradient
def dxdw_cak(u,beta,vini):
    #return x_cak(u,beta)**2 * u**(1/beta-1)/beta
    coef = ((u-vini)/(1-vini))**(1/beta)/(beta*(u-vini))
    return coef * x_cak(u,beta,vini)**2
def dxdw_aplaw(u,beta,vini):
    #return 2*u/(beta-1) * x_aplaw(u,beta)**beta
    coef = 2*(u-vini)/((beta-1)*(1-vini)**2)
    return coef * x_aplaw(u,beta,vini)**beta
def dxdw_vplaw(u,beta,vini):
    #return (x_vplaw(u,beta)-1)/(beta*u)
    return (x_vplaw(u,beta,vini)-1)/(beta*(u-vini))
def dxdw_M2005(xv,beta,vini):
    coef = 2*(u-vini)/((1-vini)**2)
    return x_M2005(xv,beta,vini)*coeff
def dxdw_expon(u,beta,vini):
    return (beta*(1-u))**-1
###
### Dictionaries of Possible Models
###
# normalized density profiles
global n,x,v
n = {'PowerLaw':        dens_plaw,\
     'PowerLaw2':       dens_dplaw,\
     'Exponential':     dens_exp,\
     'DLogistic':       dens_dlogic,\
     'LogNormal':       dens_lognorm,\
     'Normal':          dens_norm,\
     'Shell':           dens_shell,\
     'FRED':            dens_fred,\
     'Pulses':          dens_pulseslin,\
     'DampedPulses':    dens_pulsedamplin,\
     'PacketPulses':    dens_pulsegrouplin,\
     'PulsesLog':       dens_pulseslog,\
     'DampedPulsesLog': dens_pulsedamplog,\
     'PacketPulsesLog': dens_pulsegrouplog,\
     }
# normalized radial profiles
x = {'VelPlaw':   x_vplaw,\
     'AccPlaw':   x_aplaw,\
     'BetaCAK':   x_cak,\
     'Expontl':   x_expon,\
     'M2005TK':   x_M2005}
# normalized velocity fields
v = {'VelPlaw':   w_vplaw,\
     'AccPlaw':   w_aplaw,\
     'BetaCAK':   w_cak,\
     'Expontl':   w_expon,\
     'M2005TK':   w_M2005}
# normalized velocity gradients
dxdv = {'VelPlaw':   dxdw_vplaw,\
        'AccPlaw':   dxdw_aplaw,\
        'BetaCAK':   dxdw_cak,\
        'Expontl':   dxdw_expon,
        'M2005TK':   dxdw_M2005}
###
### convenience functions for static gas
###
# Gaussian profile
def gauss(w,w0,sigv):
    return exp(-0.5*((w-w0)/(w0*sigv))**2)/sqrt(2*pi*(w0*sigv)**2)
# Voigt profile
def voigt(w,w0,sigv):
    return wofz((w-w0)/(w0*sigv)).real
# derivative of a Gaussian
def dog(w,w0,var=1e-7):
    return -w/var * gauss(w,w0,var)
###
### Geometry Root-Finders
###
# difference between observed and predicted velocity for source
def diff_u0(u0,beta,vini,yi,VF):
    return yi - u0/x[VF](u0,beta,vini)*sqrt(x[VF](u0,beta,vini)**2-1)
# solve for velocity at source edge
def solve_u0(beta,vini,VF,yi):
    # check if a sign change occurs
    if diff_u0(yi,beta,vini,yi,VF) *  diff_u0(1-2**-24,beta,vini,yi,VF) < 0 :
        # if so, root-find to determine the minimum velocity
        return brentq(diff_u0,yi,1-2**-24,maxiter=100,args=(beta,vini,yi,VF))
    # if not, field is close to terminal velocity and profile is ~ zero-valued
    else:
        return 1-2**-24
# difference between observed and predicted velocity for aperture
def diff_u1(u1,beta,vini,yi,ya,VF):
    return yi - u1/x[VF](u1,beta,vini)*sqrt(x[VF](u1,beta,vini)**2-x[VF](ya,beta,vini)**2)
# solve for velocity at aperture edge
def solve_u1(ya,beta,vini,VF,yi):
    # check if a sign change occurs
    if diff_u1(ya,beta,vini,yi,ya,VF) *  diff_u1(1-2**-24,beta,vini,yi,ya,VF) < 0 :
        # if so, root-find to determine the minimum velocity
        return brentq(diff_u1,ya,1-2**-24,maxiter=100,args=(beta,vini,yi,ya,VF))
    # if not, field is close to aperture so integration stops at aperture
    else:
        return ya
# monte carlo integrator, default n_samp=10^4 gives 1% error since E~sqrt(N)
def int_mc(func,w0,w1,u,vinf,beta,vini,geo,cone,par,VF,DP,n_samp=1e4):
    intgrd = lambda w: func(w,u,vinf,beta,vini,geo,cone,par,VF,DP)
    fun_mx = brentq(lambda w: -intgrd(w),w0,w1,maxiter=100)
    w_mc   = uniform_sampler(int(n_samp))*(w1-w0)+w0
    fun_mc = uniform_sampler(int(n_samp))*(fun_mx*(1+2**-8))
    weight = (w1-w0)*fun_mx
    n_incl = count_nonzero(fun_mc<=intgrd(w_mc))
    return weight*n_incl/n_samp
###
### Non-spherical Geometries
###
# pre-calculate geometry terms which do not depend on velocity
# to speed up the integral
def precalc_geometry(incl,tO,tC,vdisk):
    geo = []
    for theta in [tO,tC]:
        # spherical cap projected as ellipse located at S
        # with horizontal axis g and vertical axis h
        h = 1.
        if theta < 2**-20 :        g = 0.
        elif theta > pi/2-2**-20 : g = 1.
        else:
            g = sin(theta)
            h *= sin(theta)
        # no inclination
        if incl < 2**-20 : S = 0.
        # with inclination
        else:
            S  = sin(incl)*cos(theta)
            h *= cos(incl)
        # coefficients for deprojected velocity -- ellipse intersection
        A = (g/h)**2-1
        B = 2*S*(g/h)**2
        C = ((S/h)**2-1)*g**2
        # store
        geo += [S,g,h,A,B,C,incl,theta]
        # if a disk term included, calculate radius and projection
        # in velocity space for comparison with deprojected velocity
        if vdisk > 1-2**-20 : geo += [1]
        elif vdisk < 2**-20 : geo += [0]
        else:                 geo += [vdisk]
        if incl < 2**-20 :    geo += [0,1]
        else:                 geo += [vdisk*cos(incl),cos(incl)**-2-1]
    return array(geo)
# compute angular length of deprojected velocity inscribed by hemisphere
# rdisk = radius of disk
# rdisk_sini = disk with inclination projection
# adisk = sin(i) projection coefficient for intersection
def hemi_inscr(u,vinf,i,t,rdisk,rdisk_sini,adisk,cone,w):
    # if w cannot project onto u
    # then return no circle
    if u > w : return 0.
    # if not inclined -- face-on
    if i <= 2**-20 :
        if cone == 'post' : return 0
        if cone == 'ante' : return 1
    # if fully inclined -- edge-on
    if i >= pi/2 - 2**-20 :
        return 0.5
    # radius of deprojected velocity from cos varrho = u/w
    rdprv = sqrt(1-(u/w)**2)
    # if disk spans all of the deprojected velocity
    if rdisk_sini > rdprv :
        #print(u,w,rdprv,rdisk_sini)
        # posterior to disk fully obstructed
        if cone == 'post' : return 0.
        # anterior to disk fully inscribed
        if cone == 'ante' : return 1.
    # otherwise, disk obstructs some of the deprojected velocity
    # hemisphere disk radius always 1
    # intersection points (pd,qd) of deprojected velocity with projected disk
    qd = sqrt((1-rdprv**2)/adisk)
    if cone == 'post' : return arccos(qd/rdprv)/pi
    else : return 1-arccos(qd/rdprv)/pi
# compute angular length of deprojected velocity inscribed by cone
# S = displacement of ellipse
# g = axis of ellipse in x coordinate
# h = axis of ellipse in y coordinate
# A,B,C = coefficients for intersection of deprojected velocity with bicone ellipse
# rdisk, adisk = radius of disk
# adisk = sin(i) projection coefficient for intersection
def cone_inscr(u,vinf,S,g,h,A,B,C,i,t,rdisk,rdisk_sini,adisk,cone,w):
    # if w cannot project onto u
    # then return no circle
    if u > w : return 0.
    # radius sin varrho of deprojected velocity
    # from cos varrho = u/w and sin^2 + cos^2 = 1
    rdprv = sqrt(1-(u/w)**2)
    # intersection points (p,q) of deprojected velocity with projected spherical cap
    C += rdprv**2
    qc = (B+array([-1,1])*sqrt(B**2-4*A*C))/(2*A)
    # pc = sqrt(rbnd**2-qc**2) # --> no need to calculate
    # unit arc length
    dl = 0.
    # if no intersection occurs, qc[0] (always smaller than qc[1]) > rdprv
    # or b^2 < 4ac, giving a NaN intersection
    # if u/w close to 1, then arc length -> 0 due to integral limit
    if qc[0] > rdprv or B**2 < 4*A*C or u/w > 1-2**-20 : return 0.
    # if deprojected velocity is fully inscribed by the ellipse
    # deprojected velocity is fully seen
    elif rdprv < h-S : return 1.
    # otherwise, full cone treatment
    else:
        # if first intersection below x-axis,
        # include full deprojected velocity minus the angle subtended by (pc,qc)
        if qc[0] < 0 and qc[0] > -rdprv : dl += 1-arccos(qc[0]/rdprv)/pi
        # if first intersection above x-axis,
        # include angle subtended by (pc,qc)
        elif qc[0] < rdprv : dl += arccos(qc[0]/rdprv)/pi
        # if second intersection occurs,
        # this is the elliptical lune projected by the other cone
        # and is the arc length of the angle subtended by (pc,qc)
        if qc[1] <= rdprv and i+t > pi/2 : dl += arccos(qc[1]/rdprv)/pi
    return dl
# compute angular length of deprojected velocity inscribed by cone + disk
# S = displacement of ellipse
# g = axis of ellipse in x coordinate
# h = axis of ellipse in y coordinate
# A,B,C = coefficients for intersection of deprojected velocity with bicone ellipse
# rdisk, adisk = radius of disk
# adisk = sin(i) projection coefficient for intersection
def cndk_inscr(u,vinf,S,g,h,A,B,C,i,t,rdisk,rdisk_sini,adisk,cone,w):
    # if w cannot project onto u
    # then return no circle
    if u > w : return 0.
    # radius of deprojected velocity from cos varrho = u/w
    rdprv = sqrt(1-(u/w)**2)
    # intersection points (p,q) of deprojected velocity with projected spherical cap
    C += rdprv**2
    qc = (B+array([-1,1])*sqrt(B**2-4*A*C))/(2*A)
    # intersection points (pd,qd) of deprojected velocity with projected disk
    qd = sqrt((rdprv**2-rdisk**2)/adisk)
    # pc = sqrt(rbnd**2-qc**2) # --> no need to calculate
    # unit arc length
    dl = 0.
    # if no intersection occurs, qc[0] (always smaller than qc[1]) > rdprv
    # or b^2 < 4ac, giving a NaN intersection
    # if u/w close to 1, then arc length -> 0 due to integral limit
    if qc[0] > rdprv or B**2 < 4*A*C or u/w > 1-2**-20 : return 0.
    # if disk present and obstructs all of the deprojected velocity
    elif rdisk_sini > rdprv and cone == 'post' : return 0.
    # if deprojected velocity is fully inscribed by the ellipse
    elif rdprv < h-S :
        # if dis is too small to obstruct deprojected velocity
        # or if cone is anterior to the disk, deprojected velocity is fully seen
        if rdisk < rdprv or cone == 'ante': return 1.
        # if disk present and large enough to obstruct some of the deprojected velocity,
        # and deprojected velocity is fully enclosed by the cone
        elif rdisk > rdprv and rdisk_sini < rdprv : return 2*arccos(qd/rdprv)/pi
    # if disk present and large enough to obstruct some of the deprojected velocity,
    # consider cones minus the arc length within cone
    # which is obstructed by disk using either
    # the first intersection (posterior -- cone or cone + lune obstructed)
    # or the second intersection (anterior -- lune is obstructed)
    elif rdisk > rdprv and rdisk > 2**-20 :
        # the cone is posterior to disk
        if cone == 'post' :
            # disk is below projected cone -- no obstruction
            if qc[0] > 0 and qc[0] > qd : dl += arccos(qc[0]/rdprv)/pi
            # disk obstructs lower part of projected cone
            elif (qc[0] > 0 and qc[0] < qd) or (qc[0] < 0 and qc[0] > -qd) : dl += arccos(qd/rdprv)/pi
            # disk obstructs middle part of projected cone
            elif qc[0] < 0 and qc[0] < -qd : dl += (2*arccos(qd/rdprv)-arccos(qc[0]/rdprv))/pi
            # projected lune from anterior cone -- unaffected by disk
            if qc[1] <= rdprv and i+t > pi/2 : dl += arccos(qc[1]/rdprv)/pi
        # if cone is anterior to disk
        elif cone == 'ante' :
            # if first intersection below x-axis,
            # include full deprojected velocity minus the angle subtended by (pc,qc)
            if qc[0] < 0 and qc[0] > -rdprv : dl += 1-arccos(qc[0]/rdprv)/pi
            # if first intersection above x-axis,
            # include angle subtended by (pc,qc)
            else : dl += arccos(qc[0]/rdprv)/pi
            # if second intersection occurs and inclination + theta > pi/2,
            # there is an elliptical lune projected by the posterior cone
            # and is the arc length of the angle subtended by (pc,qc)
            # but can be obstructed by disk
            if qc[1] < rdprv and i+t > pi/2 :
                # no obstruction if lune starts above the disk
                if qc[1] > qd or rdisk < 2**-9 : dl += arccos(qc[1]/rdprv)/pi
                # otherwise, disk determines the arc length
                elif qc[1] < qd and qd < rdprv : dl += arccos(qd/rdprv)/pi
    # if no disk or if disk radius < deprojected velocity radius, just consider the cones
    else:
        # if first intersection below x-axis,
        # include full deprojected velocity minus the angle subtended by (pc,qc)
        if qc[0] < 0 and qc[0] > -rdprv : dl += 1-arccos(qc[0]/rdprv)/pi
        # if first intersection above x-axis,
        # include angle subtended by (pc,qc)
        elif qc[0] < rdprv : dl += arccos(qc[0]/rdprv)/pi
        # if second intersection occurs,
        # this is the elliptical lune projected by the other cone
        # and is the arc length of the angle subtended by (pc,qc)
        if qc[1] <= rdprv and i+t > pi/2 : dl += arccos(qc[1]/rdprv)/pi
    return dl



from numpy import pi, inf
# dictionary of all parameter names
ProfName = {'Static':
            {'Nebular':['FluxStatic'],\
            'Absorption':['LogColumnStatic'],\
            'Resonant':['LogColumnStatic'],\
            'Fluorescent':['LogColumnStatic'],\
            'PCygni':['LogColumnStatic']},\
            'Outflow':
            {'Nebular':['FluxOutflow'],\
            'Absorption':['LogColumnOutflow'],\
            'Resonant':['LogColumnOutflow'],\
            'Fluorescent':['LogColumnOutflow'],\
            'PCygni':['LogColumnOutflow']}}
GeomName = {'Spherical':[],\
            'Hemisphere':['Inclination'],\
            'FilledCones':['Inclination','OpeningAngle'],\
            'HollowConesFixedCavity':['Inclination','OpeningAngle'],\
            'HollowCones':['Inclination','OpeningAngle','CavityAngle']}
DensName = {'PowerLaw':['PowerLawIndex'],\
            'Exponential':['DecayRate'],\
            'PowerLaw2':['PowerLawIndex1','PowerLawIndex2','Inflection'],\
            'Normal':['Peak','Width'],\
            'LogNormal':['LogPeak','LogWidth'],\
            'DLogistic':['Peak','GrowthRate'],\
            'Shell':['Radius','Width'],\
            'FRED':['RiseRate','DecayRate','Peak'],\
            'Pulses':['PulseWidth','PulseInterval','InitPeak'],\
            'DampedPulses':['DecayRate',\
                      'PulseWidth','PulseInterval','InitPeak'],\
            'PacketPulses':['PacketWidth','PacketPeak',\
                      'PulseWidth','PulseInterval','InitPeak'],\
            'PulsesLog':['PulseWidth','PulseInterval','InitPeak'],\
            'DampedPulsesLog':['DecayRate',\
                      'PulseWidth','PulseInterval','InitPeak'],\
            'PacketPulsesLog':['PacketWidth','PacketPeak',\
                      'PulseWidth','PulseInterval','InitPeak'],\
            }
# dictionary of all parameter labels
ProfLabs = {'Static':
            {'Nebular':[r'$F_s$'],\
            'Absorption':[r'$\log N_s$'],\
            'Resonant':[r'$\log N_s$'],\
            'Fluorescent':[r'$\log N_s$'],\
            'PCygni':[r'$\log N_s$']},\
            'Outflow':
            {'Nebular':[r'$F_o$'],\
            'Absorption':[r'$\log N_o$'],\
            'Resonant':[r'$\log N_o$'],\
            'Fluorescent':[r'$\log N_o$'],\
            'PCygni':[r'$\log N_o$']}}
GeomLabs = {'Spherical':[],\
            'Hemisphere':[r'$i$'],\
            'FilledCones':[r'$i$',r'$\theta_o$'],\
            'HollowConesFixedCavity':[r'$i$',r'$\theta_o$'],\
            'HollowCones':[r'$i$',r'$\theta_o$',r'$\theta_c$']}
DensLabs = {'PowerLaw':[r'$\alpha$'],\
            'Exponential':[r'$\gamma$'],\
            'PowerLaw2':[r'$\alpha_1$',r'$\alpha_2$',r'$x_1$'],\
            'Normal':[r'$x_1$',r'$\sigma_x$'],\
            'LogNormal':[r'$\log x_1$',r'$\log\sigma_x$'],\
            'DLogistic':['$x_1$','$\alpha$'],\
            'Shell':[r'$x_1$',r'$\sigma_x$'],\
            'FRED':[r'$r_1$',r'$r_2$',r'$x_1$'],\
            'Pulses':[r'$\sigma_x$',r'$\Delta x$',r'$x_0$'],\
            'DampedPulses':[r'$\gamma$',\
                      r'$\sigma_x$',r'$\Delta x$',r'$x_0$'],\
            'PacketPulses':[r'$\sigma_p$','$x_p$',\
                      r'$\sigma_x$',r'$\Delta x$',r'$x_0$'],\
            'PulsesLog':[r'$\sigma_x$',r'$\Delta \log x$',r'$\log x_0$'],\
            'DampedPulsesLog':[r'$\gamma$',\
                      r'$\sigma_x$',r'$\Delta \log x$',r'$\log x_0$'],\
            'PacketPulsesLog':[r'$\sigma_p$','$x_p$',\
                      r'$\sigma_x$',r'$\Delta \log x$',r'$\log x_0$'],\
            }
# dictionary of all parameter initial guesses
BetaPars = {'BetaCAK':[1],\
            'AccPlaw':[2],\
            'VelPlaw':[1],\
            'Expontl':[1]}
ProfPars = {'Static':
            {'Nebular':[1],\
            'Absorption':[14],\
            'Resonant':[14],\
            'Fluorescent':[14],\
            'PCygni':[14]},\
            'Outflow':
            {'Nebular':[1],\
            'Absorption':[14],\
            'Resonant':[14],\
            'Fluorescent':[14],\
            'PCygni':[14]}}
GeomPars = {'Spherical':[],\
            'Hemisphere':[pi/4],\
            'FilledCones':[pi/6,pi/6],\
            'HollowConesFixedCavity':[pi/6,pi/6],\
            'HollowCones':[pi/3,pi/4,pi/6]}
DensPars = {'PowerLaw':[2],\
            'Exponential':[0.1],\
            'PowerLaw2':[1,2,3],\
            'Normal':[3,1],\
            'LogNormal':[0.5,0.1],\
            'DLogistic':[3,1],\
            'Shell':[3,1],\
            'FRED':[1,1,3],\
            'Pulses':[0.2,1,2],\
            'DampedPulses':[0.1,0.2,1,2],\
            'PacketPulses':[1,4,0.2,1,2],\
            'PulsesLog':[0.2,0.1,0.3],\
            'DampedPulsesLog':[0.1,0.2,0.1,0.3],\
            'PacketPulsesLog':[0.1,0.5,0.2,0.1,0.3],\
            }
# dictionary of all recommended parameter bounds
BetaBounds = {'BetaCAK':[[0.1],[5]],\
            'AccPlaw':[[1.001],[5]],\
            'VelPlaw':[[0.1],[5]],\
            'Expontl':[[0.001],[5]]}
ProfBounds = {'Static':
            {'Nebular':[[0],[inf]],\
            'Absorption':[[10],[25]],\
            'Resonant':[[10],[25]],\
            'Fluorescent':[[10],[25]],\
            'PCygni':[[10],[25]]},\
            'Outflow':
            {'Nebular':[[0],[inf]],\
            'Absorption':[[10],[25]],\
            'Resonant':[[10],[25]],\
            'Fluorescent':[[10],[25]],\
            'PCygni':[[10],[25]]}}
GeomBounds = {'Spherical':[[],[]],\
            'Hemisphere':[[0],[pi/2]],\
            'FilledCones':[[0,0],[pi/2,pi/2]],\
            'HollowConesFixedCavity':[[0,0.174],[pi/2,pi/2]],\
            'HollowCones':[[0,0,0],[pi/2,pi/2,pi/2]]}
DensBounds = {'PowerLaw':[[0],[10]],\
            'Exponential':[[0],[10]],\
            'PowerLaw2':[[0,0,1],[5,10,10]],\
            'Normal':[[1,0.01],[100,10]],\
            'LogNormal':[[0,0.01],[2,1]],\
            'DLogistic':[[1,0],[100,10]],\
            'Shell':[[1,0.01],[100,10]],\
            'FRED':[[0.01,0.01,1],[100,100,100]],\
            'Pulses':[[0,0,0],[1,3,10]],\
            'DampedPulses':[[0,0,0,0],[10,1,3,10]],\
            'PacketPulses':[[0,0,0,0,0],[3,10,1,3,10]],\
            'PulsesLog':[[0,0,0],[1,1,1]],\
            'DampedPulsesLog':[[0,0,0,0],[10,1,0.5,1]],\
            'PacketPulsesLog':[[0,0,0,0],[0.7,1,1,0.5,1]]
            }
AperBounds = [[0],[1]]
DiskBounds = [[1],[inf]]
